# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)

# DatabaseId: Unique identifier of Database object.
DatabaseId = str

# Database: Database object.
class Database(ChromeTypeBase):
    def __init__(self,
                 id: Union['DatabaseId'],
                 domain: Union['str'],
                 name: Union['str'],
                 version: Union['str'],
                 ):

        self.id = id
        self.domain = domain
        self.name = name
        self.version = version


# Error: Database error.
class Error(ChromeTypeBase):
    def __init__(self,
                 message: Union['str'],
                 code: Union['int'],
                 ):

        self.message = message
        self.code = code


class Database(PayloadMixin):
    """ 
    """
    @classmethod
    def disable(cls):
        """Disables database tracking, prevents database events from being sent to the client.
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def enable(cls):
        """Enables database tracking, database events will now be delivered to the client.
        """
        return (
            cls.build_send_payload("enable", {
            }),
            None
        )

    @classmethod
    def executeSQL(cls,
                   databaseId: Union['DatabaseId'],
                   query: Union['str'],
                   ):
        """
        :param databaseId: 
        :type databaseId: DatabaseId
        :param query: 
        :type query: str
        """
        return (
            cls.build_send_payload("executeSQL", {
                "databaseId": databaseId,
                "query": query,
            }),
            cls.convert_payload({
                "columnNames": {
                    "class": [],
                    "optional": True
                },
                "values": {
                    "class": [],
                    "optional": True
                },
                "sqlError": {
                    "class": Error,
                    "optional": True
                },
            })
        )

    @classmethod
    def getDatabaseTableNames(cls,
                              databaseId: Union['DatabaseId'],
                              ):
        """
        :param databaseId: 
        :type databaseId: DatabaseId
        """
        return (
            cls.build_send_payload("getDatabaseTableNames", {
                "databaseId": databaseId,
            }),
            cls.convert_payload({
                "tableNames": {
                    "class": [],
                    "optional": False
                },
            })
        )



class AddDatabaseEvent(BaseEvent):

    js_name = 'Database.addDatabase'
    hashable = ['databaseId']
    is_hashable = True

    def __init__(self,
                 database: Union['Database', dict],
                 ):
        if isinstance(database, dict):
            database = Database(**database)
        self.database = database

    @classmethod
    def build_hash(cls, databaseId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h



# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)
from chromewhip.protocol import dom as DOM
from chromewhip.protocol import page as Page
from chromewhip.protocol import runtime as Runtime

# ScreenOrientation: Screen orientation.
class ScreenOrientation(ChromeTypeBase):
    def __init__(self,
                 type: Union['str'],
                 angle: Union['int'],
                 ):

        self.type = type
        self.angle = angle


# VirtualTimePolicy: advance: If the scheduler runs out of immediate work, the virtual time base may fast forward toallow the next delayed task (if any) to run; pause: The virtual time base may not advance;pauseIfNetworkFetchesPending: The virtual time base may not advance if there are any pendingresource fetches.
VirtualTimePolicy = str

class Emulation(PayloadMixin):
    """ This domain emulates different environments for the page.
    """
    @classmethod
    def canEmulate(cls):
        """Tells whether emulation is supported.
        """
        return (
            cls.build_send_payload("canEmulate", {
            }),
            cls.convert_payload({
                "result": {
                    "class": bool,
                    "optional": False
                },
            })
        )

    @classmethod
    def clearDeviceMetricsOverride(cls):
        """Clears the overriden device metrics.
        """
        return (
            cls.build_send_payload("clearDeviceMetricsOverride", {
            }),
            None
        )

    @classmethod
    def clearGeolocationOverride(cls):
        """Clears the overriden Geolocation Position and Error.
        """
        return (
            cls.build_send_payload("clearGeolocationOverride", {
            }),
            None
        )

    @classmethod
    def resetPageScaleFactor(cls):
        """Requests that page scale factor is reset to initial values.
        """
        return (
            cls.build_send_payload("resetPageScaleFactor", {
            }),
            None
        )

    @classmethod
    def setFocusEmulationEnabled(cls,
                                 enabled: Union['bool'],
                                 ):
        """Enables or disables simulating a focused and active page.
        :param enabled: Whether to enable to disable focus emulation.
        :type enabled: bool
        """
        return (
            cls.build_send_payload("setFocusEmulationEnabled", {
                "enabled": enabled,
            }),
            None
        )

    @classmethod
    def setCPUThrottlingRate(cls,
                             rate: Union['float'],
                             ):
        """Enables CPU throttling to emulate slow CPUs.
        :param rate: Throttling rate as a slowdown factor (1 is no throttle, 2 is 2x slowdown, etc).
        :type rate: float
        """
        return (
            cls.build_send_payload("setCPUThrottlingRate", {
                "rate": rate,
            }),
            None
        )

    @classmethod
    def setDefaultBackgroundColorOverride(cls,
                                          color: Optional['DOM.RGBA'] = None,
                                          ):
        """Sets or clears an override of the default background color of the frame. This override is used
if the content does not specify one.
        :param color: RGBA of the default background color. If not specified, any existing override will be
cleared.
        :type color: DOM.RGBA
        """
        return (
            cls.build_send_payload("setDefaultBackgroundColorOverride", {
                "color": color,
            }),
            None
        )

    @classmethod
    def setDeviceMetricsOverride(cls,
                                 width: Union['int'],
                                 height: Union['int'],
                                 deviceScaleFactor: Union['float'],
                                 mobile: Union['bool'],
                                 scale: Optional['float'] = None,
                                 screenWidth: Optional['int'] = None,
                                 screenHeight: Optional['int'] = None,
                                 positionX: Optional['int'] = None,
                                 positionY: Optional['int'] = None,
                                 dontSetVisibleSize: Optional['bool'] = None,
                                 screenOrientation: Optional['ScreenOrientation'] = None,
                                 viewport: Optional['Page.Viewport'] = None,
                                 ):
        """Overrides the values of device screen dimensions (window.screen.width, window.screen.height,
window.innerWidth, window.innerHeight, and "device-width"/"device-height"-related CSS media
query results).
        :param width: Overriding width value in pixels (minimum 0, maximum 10000000). 0 disables the override.
        :type width: int
        :param height: Overriding height value in pixels (minimum 0, maximum 10000000). 0 disables the override.
        :type height: int
        :param deviceScaleFactor: Overriding device scale factor value. 0 disables the override.
        :type deviceScaleFactor: float
        :param mobile: Whether to emulate mobile device. This includes viewport meta tag, overlay scrollbars, text
autosizing and more.
        :type mobile: bool
        :param scale: Scale to apply to resulting view image.
        :type scale: float
        :param screenWidth: Overriding screen width value in pixels (minimum 0, maximum 10000000).
        :type screenWidth: int
        :param screenHeight: Overriding screen height value in pixels (minimum 0, maximum 10000000).
        :type screenHeight: int
        :param positionX: Overriding view X position on screen in pixels (minimum 0, maximum 10000000).
        :type positionX: int
        :param positionY: Overriding view Y position on screen in pixels (minimum 0, maximum 10000000).
        :type positionY: int
        :param dontSetVisibleSize: Do not set visible view size, rely upon explicit setVisibleSize call.
        :type dontSetVisibleSize: bool
        :param screenOrientation: Screen orientation override.
        :type screenOrientation: ScreenOrientation
        :param viewport: If set, the visible area of the page will be overridden to this viewport. This viewport
change is not observed by the page, e.g. viewport-relative elements do not change positions.
        :type viewport: Page.Viewport
        """
        return (
            cls.build_send_payload("setDeviceMetricsOverride", {
                "width": width,
                "height": height,
                "deviceScaleFactor": deviceScaleFactor,
                "mobile": mobile,
                "scale": scale,
                "screenWidth": screenWidth,
                "screenHeight": screenHeight,
                "positionX": positionX,
                "positionY": positionY,
                "dontSetVisibleSize": dontSetVisibleSize,
                "screenOrientation": screenOrientation,
                "viewport": viewport,
            }),
            None
        )

    @classmethod
    def setScrollbarsHidden(cls,
                            hidden: Union['bool'],
                            ):
        """
        :param hidden: Whether scrollbars should be always hidden.
        :type hidden: bool
        """
        return (
            cls.build_send_payload("setScrollbarsHidden", {
                "hidden": hidden,
            }),
            None
        )

    @classmethod
    def setDocumentCookieDisabled(cls,
                                  disabled: Union['bool'],
                                  ):
        """
        :param disabled: Whether document.coookie API should be disabled.
        :type disabled: bool
        """
        return (
            cls.build_send_payload("setDocumentCookieDisabled", {
                "disabled": disabled,
            }),
            None
        )

    @classmethod
    def setEmitTouchEventsForMouse(cls,
                                   enabled: Union['bool'],
                                   configuration: Optional['str'] = None,
                                   ):
        """
        :param enabled: Whether touch emulation based on mouse input should be enabled.
        :type enabled: bool
        :param configuration: Touch/gesture events configuration. Default: current platform.
        :type configuration: str
        """
        return (
            cls.build_send_payload("setEmitTouchEventsForMouse", {
                "enabled": enabled,
                "configuration": configuration,
            }),
            None
        )

    @classmethod
    def setEmulatedMedia(cls,
                         media: Union['str'],
                         ):
        """Emulates the given media for CSS media queries.
        :param media: Media type to emulate. Empty string disables the override.
        :type media: str
        """
        return (
            cls.build_send_payload("setEmulatedMedia", {
                "media": media,
            }),
            None
        )

    @classmethod
    def setGeolocationOverride(cls,
                               latitude: Optional['float'] = None,
                               longitude: Optional['float'] = None,
                               accuracy: Optional['float'] = None,
                               ):
        """Overrides the Geolocation Position or Error. Omitting any of the parameters emulates position
unavailable.
        :param latitude: Mock latitude
        :type latitude: float
        :param longitude: Mock longitude
        :type longitude: float
        :param accuracy: Mock accuracy
        :type accuracy: float
        """
        return (
            cls.build_send_payload("setGeolocationOverride", {
                "latitude": latitude,
                "longitude": longitude,
                "accuracy": accuracy,
            }),
            None
        )

    @classmethod
    def setNavigatorOverrides(cls,
                              platform: Union['str'],
                              ):
        """Overrides value returned by the javascript navigator object.
        :param platform: The platform navigator.platform should return.
        :type platform: str
        """
        return (
            cls.build_send_payload("setNavigatorOverrides", {
                "platform": platform,
            }),
            None
        )

    @classmethod
    def setPageScaleFactor(cls,
                           pageScaleFactor: Union['float'],
                           ):
        """Sets a specified page scale factor.
        :param pageScaleFactor: Page scale factor.
        :type pageScaleFactor: float
        """
        return (
            cls.build_send_payload("setPageScaleFactor", {
                "pageScaleFactor": pageScaleFactor,
            }),
            None
        )

    @classmethod
    def setScriptExecutionDisabled(cls,
                                   value: Union['bool'],
                                   ):
        """Switches script execution in the page.
        :param value: Whether script execution should be disabled in the page.
        :type value: bool
        """
        return (
            cls.build_send_payload("setScriptExecutionDisabled", {
                "value": value,
            }),
            None
        )

    @classmethod
    def setTouchEmulationEnabled(cls,
                                 enabled: Union['bool'],
                                 maxTouchPoints: Optional['int'] = None,
                                 ):
        """Enables touch on platforms which do not support them.
        :param enabled: Whether the touch event emulation should be enabled.
        :type enabled: bool
        :param maxTouchPoints: Maximum touch points supported. Defaults to one.
        :type maxTouchPoints: int
        """
        return (
            cls.build_send_payload("setTouchEmulationEnabled", {
                "enabled": enabled,
                "maxTouchPoints": maxTouchPoints,
            }),
            None
        )

    @classmethod
    def setVirtualTimePolicy(cls,
                             policy: Union['VirtualTimePolicy'],
                             budget: Optional['float'] = None,
                             maxVirtualTimeTaskStarvationCount: Optional['int'] = None,
                             waitForNavigation: Optional['bool'] = None,
                             initialVirtualTime: Optional['Network.TimeSinceEpoch'] = None,
                             ):
        """Turns on virtual time for all frames (replacing real-time with a synthetic time source) and sets
the current virtual time policy.  Note this supersedes any previous time budget.
        :param policy: 
        :type policy: VirtualTimePolicy
        :param budget: If set, after this many virtual milliseconds have elapsed virtual time will be paused and a
virtualTimeBudgetExpired event is sent.
        :type budget: float
        :param maxVirtualTimeTaskStarvationCount: If set this specifies the maximum number of tasks that can be run before virtual is forced
forwards to prevent deadlock.
        :type maxVirtualTimeTaskStarvationCount: int
        :param waitForNavigation: If set the virtual time policy change should be deferred until any frame starts navigating.
Note any previous deferred policy change is superseded.
        :type waitForNavigation: bool
        :param initialVirtualTime: If set, base::Time::Now will be overriden to initially return this value.
        :type initialVirtualTime: Network.TimeSinceEpoch
        """
        return (
            cls.build_send_payload("setVirtualTimePolicy", {
                "policy": policy,
                "budget": budget,
                "maxVirtualTimeTaskStarvationCount": maxVirtualTimeTaskStarvationCount,
                "waitForNavigation": waitForNavigation,
                "initialVirtualTime": initialVirtualTime,
            }),
            cls.convert_payload({
                "virtualTimeTicksBase": {
                    "class": float,
                    "optional": False
                },
            })
        )

    @classmethod
    def setVisibleSize(cls,
                       width: Union['int'],
                       height: Union['int'],
                       ):
        """Resizes the frame/viewport of the page. Note that this does not affect the frame's container
(e.g. browser window). Can be used to produce screenshots of the specified size. Not supported
on Android.
        :param width: Frame width (DIP).
        :type width: int
        :param height: Frame height (DIP).
        :type height: int
        """
        return (
            cls.build_send_payload("setVisibleSize", {
                "width": width,
                "height": height,
            }),
            None
        )

    @classmethod
    def setUserAgentOverride(cls,
                             userAgent: Union['str'],
                             acceptLanguage: Optional['str'] = None,
                             platform: Optional['str'] = None,
                             ):
        """Allows overriding user agent with the given string.
        :param userAgent: User agent to use.
        :type userAgent: str
        :param acceptLanguage: Browser langugage to emulate.
        :type acceptLanguage: str
        :param platform: The platform navigator.platform should return.
        :type platform: str
        """
        return (
            cls.build_send_payload("setUserAgentOverride", {
                "userAgent": userAgent,
                "acceptLanguage": acceptLanguage,
                "platform": platform,
            }),
            None
        )



class VirtualTimeAdvancedEvent(BaseEvent):

    js_name = 'Emulation.virtualTimeAdvanced'
    hashable = []
    is_hashable = False

    def __init__(self,
                 virtualTimeElapsed: Union['float', dict],
                 ):
        if isinstance(virtualTimeElapsed, dict):
            virtualTimeElapsed = float(**virtualTimeElapsed)
        self.virtualTimeElapsed = virtualTimeElapsed

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class VirtualTimeBudgetExpiredEvent(BaseEvent):

    js_name = 'Emulation.virtualTimeBudgetExpired'
    hashable = []
    is_hashable = False

    def __init__(self):
        pass

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class VirtualTimePausedEvent(BaseEvent):

    js_name = 'Emulation.virtualTimePaused'
    hashable = []
    is_hashable = False

    def __init__(self,
                 virtualTimeElapsed: Union['float', dict],
                 ):
        if isinstance(virtualTimeElapsed, dict):
            virtualTimeElapsed = float(**virtualTimeElapsed)
        self.virtualTimeElapsed = virtualTimeElapsed

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')



# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)
from chromewhip.protocol import network as Network
from chromewhip.protocol import io as IO
from chromewhip.protocol import page as Page

# RequestId: Unique request identifier.
RequestId = str

# RequestStage: Stages of the request to handle. Request will intercept before the request issent. Response will intercept after the response is received (but before responsebody is received.
RequestStage = str

# RequestPattern: 
class RequestPattern(ChromeTypeBase):
    def __init__(self,
                 urlPattern: Optional['str'] = None,
                 resourceType: Optional['Network.ResourceType'] = None,
                 requestStage: Optional['RequestStage'] = None,
                 ):

        self.urlPattern = urlPattern
        self.resourceType = resourceType
        self.requestStage = requestStage


# HeaderEntry: Response HTTP header entry
class HeaderEntry(ChromeTypeBase):
    def __init__(self,
                 name: Union['str'],
                 value: Union['str'],
                 ):

        self.name = name
        self.value = value


# AuthChallenge: Authorization challenge for HTTP status code 401 or 407.
class AuthChallenge(ChromeTypeBase):
    def __init__(self,
                 origin: Union['str'],
                 scheme: Union['str'],
                 realm: Union['str'],
                 source: Optional['str'] = None,
                 ):

        self.source = source
        self.origin = origin
        self.scheme = scheme
        self.realm = realm


# AuthChallengeResponse: Response to an AuthChallenge.
class AuthChallengeResponse(ChromeTypeBase):
    def __init__(self,
                 response: Union['str'],
                 username: Optional['str'] = None,
                 password: Optional['str'] = None,
                 ):

        self.response = response
        self.username = username
        self.password = password


class Fetch(PayloadMixin):
    """ A domain for letting clients substitute browser's network layer with client code.
    """
    @classmethod
    def disable(cls):
        """Disables the fetch domain.
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def enable(cls,
               patterns: Optional['[RequestPattern]'] = None,
               handleAuthRequests: Optional['bool'] = None,
               ):
        """Enables issuing of requestPaused events. A request will be paused until client
calls one of failRequest, fulfillRequest or continueRequest/continueWithAuth.
        :param patterns: If specified, only requests matching any of these patterns will produce
fetchRequested event and will be paused until clients response. If not set,
all requests will be affected.
        :type patterns: [RequestPattern]
        :param handleAuthRequests: If true, authRequired events will be issued and requests will be paused
expecting a call to continueWithAuth.
        :type handleAuthRequests: bool
        """
        return (
            cls.build_send_payload("enable", {
                "patterns": patterns,
                "handleAuthRequests": handleAuthRequests,
            }),
            None
        )

    @classmethod
    def failRequest(cls,
                    requestId: Union['RequestId'],
                    errorReason: Union['Network.ErrorReason'],
                    ):
        """Causes the request to fail with specified reason.
        :param requestId: An id the client received in requestPaused event.
        :type requestId: RequestId
        :param errorReason: Causes the request to fail with the given reason.
        :type errorReason: Network.ErrorReason
        """
        return (
            cls.build_send_payload("failRequest", {
                "requestId": requestId,
                "errorReason": errorReason,
            }),
            None
        )

    @classmethod
    def fulfillRequest(cls,
                       requestId: Union['RequestId'],
                       responseCode: Union['int'],
                       responseHeaders: Union['[HeaderEntry]'],
                       body: Optional['str'] = None,
                       responsePhrase: Optional['str'] = None,
                       ):
        """Provides response to the request.
        :param requestId: An id the client received in requestPaused event.
        :type requestId: RequestId
        :param responseCode: An HTTP response code.
        :type responseCode: int
        :param responseHeaders: Response headers.
        :type responseHeaders: [HeaderEntry]
        :param body: A response body.
        :type body: str
        :param responsePhrase: A textual representation of responseCode.
If absent, a standard phrase mathcing responseCode is used.
        :type responsePhrase: str
        """
        return (
            cls.build_send_payload("fulfillRequest", {
                "requestId": requestId,
                "responseCode": responseCode,
                "responseHeaders": responseHeaders,
                "body": body,
                "responsePhrase": responsePhrase,
            }),
            None
        )

    @classmethod
    def continueRequest(cls,
                        requestId: Union['RequestId'],
                        url: Optional['str'] = None,
                        method: Optional['str'] = None,
                        postData: Optional['str'] = None,
                        headers: Optional['[HeaderEntry]'] = None,
                        ):
        """Continues the request, optionally modifying some of its parameters.
        :param requestId: An id the client received in requestPaused event.
        :type requestId: RequestId
        :param url: If set, the request url will be modified in a way that's not observable by page.
        :type url: str
        :param method: If set, the request method is overridden.
        :type method: str
        :param postData: If set, overrides the post data in the request.
        :type postData: str
        :param headers: If set, overrides the request headrts.
        :type headers: [HeaderEntry]
        """
        return (
            cls.build_send_payload("continueRequest", {
                "requestId": requestId,
                "url": url,
                "method": method,
                "postData": postData,
                "headers": headers,
            }),
            None
        )

    @classmethod
    def continueWithAuth(cls,
                         requestId: Union['RequestId'],
                         authChallengeResponse: Union['AuthChallengeResponse'],
                         ):
        """Continues a request supplying authChallengeResponse following authRequired event.
        :param requestId: An id the client received in authRequired event.
        :type requestId: RequestId
        :param authChallengeResponse: Response to  with an authChallenge.
        :type authChallengeResponse: AuthChallengeResponse
        """
        return (
            cls.build_send_payload("continueWithAuth", {
                "requestId": requestId,
                "authChallengeResponse": authChallengeResponse,
            }),
            None
        )

    @classmethod
    def getResponseBody(cls,
                        requestId: Union['RequestId'],
                        ):
        """Causes the body of the response to be received from the server and
returned as a single string. May only be issued for a request that
is paused in the Response stage and is mutually exclusive with
takeResponseBodyForInterceptionAsStream. Calling other methods that
affect the request or disabling fetch domain before body is received
results in an undefined behavior.
        :param requestId: Identifier for the intercepted request to get body for.
        :type requestId: RequestId
        """
        return (
            cls.build_send_payload("getResponseBody", {
                "requestId": requestId,
            }),
            cls.convert_payload({
                "body": {
                    "class": str,
                    "optional": False
                },
                "base64Encoded": {
                    "class": bool,
                    "optional": False
                },
            })
        )

    @classmethod
    def takeResponseBodyAsStream(cls,
                                 requestId: Union['RequestId'],
                                 ):
        """Returns a handle to the stream representing the response body.
The request must be paused in the HeadersReceived stage.
Note that after this command the request can't be continued
as is -- client either needs to cancel it or to provide the
response body.
The stream only supports sequential read, IO.read will fail if the position
is specified.
This method is mutually exclusive with getResponseBody.
Calling other methods that affect the request or disabling fetch
domain before body is received results in an undefined behavior.
        :param requestId: 
        :type requestId: RequestId
        """
        return (
            cls.build_send_payload("takeResponseBodyAsStream", {
                "requestId": requestId,
            }),
            cls.convert_payload({
                "stream": {
                    "class": IO.StreamHandle,
                    "optional": False
                },
            })
        )



class RequestPausedEvent(BaseEvent):

    js_name = 'Fetch.requestPaused'
    hashable = ['requestId', 'frameId']
    is_hashable = True

    def __init__(self,
                 requestId: Union['RequestId', dict],
                 request: Union['Network.Request', dict],
                 frameId: Union['Page.FrameId', dict],
                 resourceType: Union['Network.ResourceType', dict],
                 responseErrorReason: Union['Network.ErrorReason', dict, None] = None,
                 responseStatusCode: Union['int', dict, None] = None,
                 responseHeaders: Union['[HeaderEntry]', dict, None] = None,
                 ):
        if isinstance(requestId, dict):
            requestId = RequestId(**requestId)
        self.requestId = requestId
        if isinstance(request, dict):
            request = Network.Request(**request)
        self.request = request
        if isinstance(frameId, dict):
            frameId = Page.FrameId(**frameId)
        self.frameId = frameId
        if isinstance(resourceType, dict):
            resourceType = Network.ResourceType(**resourceType)
        self.resourceType = resourceType
        if isinstance(responseErrorReason, dict):
            responseErrorReason = Network.ErrorReason(**responseErrorReason)
        self.responseErrorReason = responseErrorReason
        if isinstance(responseStatusCode, dict):
            responseStatusCode = int(**responseStatusCode)
        self.responseStatusCode = responseStatusCode
        if isinstance(responseHeaders, dict):
            responseHeaders = [HeaderEntry](**responseHeaders)
        self.responseHeaders = responseHeaders

    @classmethod
    def build_hash(cls, requestId, frameId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class AuthRequiredEvent(BaseEvent):

    js_name = 'Fetch.authRequired'
    hashable = ['requestId', 'frameId']
    is_hashable = True

    def __init__(self,
                 requestId: Union['RequestId', dict],
                 request: Union['Network.Request', dict],
                 frameId: Union['Page.FrameId', dict],
                 resourceType: Union['Network.ResourceType', dict],
                 authChallenge: Union['AuthChallenge', dict],
                 ):
        if isinstance(requestId, dict):
            requestId = RequestId(**requestId)
        self.requestId = requestId
        if isinstance(request, dict):
            request = Network.Request(**request)
        self.request = request
        if isinstance(frameId, dict):
            frameId = Page.FrameId(**frameId)
        self.frameId = frameId
        if isinstance(resourceType, dict):
            resourceType = Network.ResourceType(**resourceType)
        self.resourceType = resourceType
        if isinstance(authChallenge, dict):
            authChallenge = AuthChallenge(**authChallenge)
        self.authChallenge = authChallenge

    @classmethod
    def build_hash(cls, requestId, frameId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h



# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)

# CertificateId: An internal certificate ID value.
CertificateId = int

# MixedContentType: A description of mixed content (HTTP resources on HTTPS pages), as defined byhttps://www.w3.org/TR/mixed-content/#categories
MixedContentType = str

# SecurityState: The security level of a page or resource.
SecurityState = str

# SecurityStateExplanation: An explanation of an factor contributing to the security state.
class SecurityStateExplanation(ChromeTypeBase):
    def __init__(self,
                 securityState: Union['SecurityState'],
                 title: Union['str'],
                 summary: Union['str'],
                 description: Union['str'],
                 mixedContentType: Union['MixedContentType'],
                 certificate: Union['[]'],
                 recommendations: Optional['[]'] = None,
                 ):

        self.securityState = securityState
        self.title = title
        self.summary = summary
        self.description = description
        self.mixedContentType = mixedContentType
        self.certificate = certificate
        self.recommendations = recommendations


# InsecureContentStatus: Information about insecure content on the page.
class InsecureContentStatus(ChromeTypeBase):
    def __init__(self,
                 ranMixedContent: Union['bool'],
                 displayedMixedContent: Union['bool'],
                 containedMixedForm: Union['bool'],
                 ranContentWithCertErrors: Union['bool'],
                 displayedContentWithCertErrors: Union['bool'],
                 ranInsecureContentStyle: Union['SecurityState'],
                 displayedInsecureContentStyle: Union['SecurityState'],
                 ):

        self.ranMixedContent = ranMixedContent
        self.displayedMixedContent = displayedMixedContent
        self.containedMixedForm = containedMixedForm
        self.ranContentWithCertErrors = ranContentWithCertErrors
        self.displayedContentWithCertErrors = displayedContentWithCertErrors
        self.ranInsecureContentStyle = ranInsecureContentStyle
        self.displayedInsecureContentStyle = displayedInsecureContentStyle


# CertificateErrorAction: The action to take when a certificate error occurs. continue will continue processing therequest and cancel will cancel the request.
CertificateErrorAction = str

class Security(PayloadMixin):
    """ Security
    """
    @classmethod
    def disable(cls):
        """Disables tracking security state changes.
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def enable(cls):
        """Enables tracking security state changes.
        """
        return (
            cls.build_send_payload("enable", {
            }),
            None
        )

    @classmethod
    def setIgnoreCertificateErrors(cls,
                                   ignore: Union['bool'],
                                   ):
        """Enable/disable whether all certificate errors should be ignored.
        :param ignore: If true, all certificate errors will be ignored.
        :type ignore: bool
        """
        return (
            cls.build_send_payload("setIgnoreCertificateErrors", {
                "ignore": ignore,
            }),
            None
        )

    @classmethod
    def handleCertificateError(cls,
                               eventId: Union['int'],
                               action: Union['CertificateErrorAction'],
                               ):
        """Handles a certificate error that fired a certificateError event.
        :param eventId: The ID of the event.
        :type eventId: int
        :param action: The action to take on the certificate error.
        :type action: CertificateErrorAction
        """
        return (
            cls.build_send_payload("handleCertificateError", {
                "eventId": eventId,
                "action": action,
            }),
            None
        )

    @classmethod
    def setOverrideCertificateErrors(cls,
                                     override: Union['bool'],
                                     ):
        """Enable/disable overriding certificate errors. If enabled, all certificate error events need to
be handled by the DevTools client and should be answered with `handleCertificateError` commands.
        :param override: If true, certificate errors will be overridden.
        :type override: bool
        """
        return (
            cls.build_send_payload("setOverrideCertificateErrors", {
                "override": override,
            }),
            None
        )



class CertificateErrorEvent(BaseEvent):

    js_name = 'Security.certificateError'
    hashable = ['eventId']
    is_hashable = True

    def __init__(self,
                 eventId: Union['int', dict],
                 errorType: Union['str', dict],
                 requestURL: Union['str', dict],
                 ):
        if isinstance(eventId, dict):
            eventId = int(**eventId)
        self.eventId = eventId
        if isinstance(errorType, dict):
            errorType = str(**errorType)
        self.errorType = errorType
        if isinstance(requestURL, dict):
            requestURL = str(**requestURL)
        self.requestURL = requestURL

    @classmethod
    def build_hash(cls, eventId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class SecurityStateChangedEvent(BaseEvent):

    js_name = 'Security.securityStateChanged'
    hashable = []
    is_hashable = False

    def __init__(self,
                 securityState: Union['SecurityState', dict],
                 schemeIsCryptographic: Union['bool', dict],
                 explanations: Union['[SecurityStateExplanation]', dict],
                 insecureContentStatus: Union['InsecureContentStatus', dict],
                 summary: Union['str', dict, None] = None,
                 ):
        if isinstance(securityState, dict):
            securityState = SecurityState(**securityState)
        self.securityState = securityState
        if isinstance(schemeIsCryptographic, dict):
            schemeIsCryptographic = bool(**schemeIsCryptographic)
        self.schemeIsCryptographic = schemeIsCryptographic
        if isinstance(explanations, dict):
            explanations = [SecurityStateExplanation](**explanations)
        self.explanations = explanations
        if isinstance(insecureContentStatus, dict):
            insecureContentStatus = InsecureContentStatus(**insecureContentStatus)
        self.insecureContentStatus = insecureContentStatus
        if isinstance(summary, dict):
            summary = str(**summary)
        self.summary = summary

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')



# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)

# CacheId: Unique identifier of the Cache object.
CacheId = str

# CachedResponseType: type of HTTP response cached
CachedResponseType = str

# DataEntry: Data entry.
class DataEntry(ChromeTypeBase):
    def __init__(self,
                 requestURL: Union['str'],
                 requestMethod: Union['str'],
                 requestHeaders: Union['[Header]'],
                 responseTime: Union['float'],
                 responseStatus: Union['int'],
                 responseStatusText: Union['str'],
                 responseType: Union['CachedResponseType'],
                 responseHeaders: Union['[Header]'],
                 ):

        self.requestURL = requestURL
        self.requestMethod = requestMethod
        self.requestHeaders = requestHeaders
        self.responseTime = responseTime
        self.responseStatus = responseStatus
        self.responseStatusText = responseStatusText
        self.responseType = responseType
        self.responseHeaders = responseHeaders


# Cache: Cache identifier.
class Cache(ChromeTypeBase):
    def __init__(self,
                 cacheId: Union['CacheId'],
                 securityOrigin: Union['str'],
                 cacheName: Union['str'],
                 ):

        self.cacheId = cacheId
        self.securityOrigin = securityOrigin
        self.cacheName = cacheName


# Header: 
class Header(ChromeTypeBase):
    def __init__(self,
                 name: Union['str'],
                 value: Union['str'],
                 ):

        self.name = name
        self.value = value


# CachedResponse: Cached response
class CachedResponse(ChromeTypeBase):
    def __init__(self,
                 body: Union['str'],
                 ):

        self.body = body


class CacheStorage(PayloadMixin):
    """ 
    """
    @classmethod
    def deleteCache(cls,
                    cacheId: Union['CacheId'],
                    ):
        """Deletes a cache.
        :param cacheId: Id of cache for deletion.
        :type cacheId: CacheId
        """
        return (
            cls.build_send_payload("deleteCache", {
                "cacheId": cacheId,
            }),
            None
        )

    @classmethod
    def deleteEntry(cls,
                    cacheId: Union['CacheId'],
                    request: Union['str'],
                    ):
        """Deletes a cache entry.
        :param cacheId: Id of cache where the entry will be deleted.
        :type cacheId: CacheId
        :param request: URL spec of the request.
        :type request: str
        """
        return (
            cls.build_send_payload("deleteEntry", {
                "cacheId": cacheId,
                "request": request,
            }),
            None
        )

    @classmethod
    def requestCacheNames(cls,
                          securityOrigin: Union['str'],
                          ):
        """Requests cache names.
        :param securityOrigin: Security origin.
        :type securityOrigin: str
        """
        return (
            cls.build_send_payload("requestCacheNames", {
                "securityOrigin": securityOrigin,
            }),
            cls.convert_payload({
                "caches": {
                    "class": [Cache],
                    "optional": False
                },
            })
        )

    @classmethod
    def requestCachedResponse(cls,
                              cacheId: Union['CacheId'],
                              requestURL: Union['str'],
                              ):
        """Fetches cache entry.
        :param cacheId: Id of cache that contains the entry.
        :type cacheId: CacheId
        :param requestURL: URL spec of the request.
        :type requestURL: str
        """
        return (
            cls.build_send_payload("requestCachedResponse", {
                "cacheId": cacheId,
                "requestURL": requestURL,
            }),
            cls.convert_payload({
                "response": {
                    "class": CachedResponse,
                    "optional": False
                },
            })
        )

    @classmethod
    def requestEntries(cls,
                       cacheId: Union['CacheId'],
                       skipCount: Union['int'],
                       pageSize: Union['int'],
                       pathFilter: Optional['str'] = None,
                       ):
        """Requests data from cache.
        :param cacheId: ID of cache to get entries from.
        :type cacheId: CacheId
        :param skipCount: Number of records to skip.
        :type skipCount: int
        :param pageSize: Number of records to fetch.
        :type pageSize: int
        :param pathFilter: If present, only return the entries containing this substring in the path
        :type pathFilter: str
        """
        return (
            cls.build_send_payload("requestEntries", {
                "cacheId": cacheId,
                "skipCount": skipCount,
                "pageSize": pageSize,
                "pathFilter": pathFilter,
            }),
            cls.convert_payload({
                "cacheDataEntries": {
                    "class": [DataEntry],
                    "optional": False
                },
                "hasMore": {
                    "class": bool,
                    "optional": False
                },
            })
        )




# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)
from chromewhip.protocol import dom as DOM

# AXNodeId: Unique accessibility node identifier.
AXNodeId = str

# AXValueType: Enum of possible property types.
AXValueType = str

# AXValueSourceType: Enum of possible property sources.
AXValueSourceType = str

# AXValueNativeSourceType: Enum of possible native property sources (as a subtype of a particular AXValueSourceType).
AXValueNativeSourceType = str

# AXValueSource: A single source for a computed AX property.
class AXValueSource(ChromeTypeBase):
    def __init__(self,
                 type: Union['AXValueSourceType'],
                 value: Optional['AXValue'] = None,
                 attribute: Optional['str'] = None,
                 attributeValue: Optional['AXValue'] = None,
                 superseded: Optional['bool'] = None,
                 nativeSource: Optional['AXValueNativeSourceType'] = None,
                 nativeSourceValue: Optional['AXValue'] = None,
                 invalid: Optional['bool'] = None,
                 invalidReason: Optional['str'] = None,
                 ):

        self.type = type
        self.value = value
        self.attribute = attribute
        self.attributeValue = attributeValue
        self.superseded = superseded
        self.nativeSource = nativeSource
        self.nativeSourceValue = nativeSourceValue
        self.invalid = invalid
        self.invalidReason = invalidReason


# AXRelatedNode: 
class AXRelatedNode(ChromeTypeBase):
    def __init__(self,
                 backendDOMNodeId: Union['DOM.BackendNodeId'],
                 idref: Optional['str'] = None,
                 text: Optional['str'] = None,
                 ):

        self.backendDOMNodeId = backendDOMNodeId
        self.idref = idref
        self.text = text


# AXProperty: 
class AXProperty(ChromeTypeBase):
    def __init__(self,
                 name: Union['AXPropertyName'],
                 value: Union['AXValue'],
                 ):

        self.name = name
        self.value = value


# AXValue: A single computed AX property.
class AXValue(ChromeTypeBase):
    def __init__(self,
                 type: Union['AXValueType'],
                 value: Optional['Any'] = None,
                 relatedNodes: Optional['[AXRelatedNode]'] = None,
                 sources: Optional['[AXValueSource]'] = None,
                 ):

        self.type = type
        self.value = value
        self.relatedNodes = relatedNodes
        self.sources = sources


# AXPropertyName: Values of AXProperty name: from 'busy' to 'roledescription' - states which apply to every AXnode, from 'live' to 'root' - attributes which apply to nodes in live regions, from'autocomplete' to 'valuetext' - attributes which apply to widgets, from 'checked' to 'selected'- states which apply to widgets, from 'activedescendant' to 'owns' - relationships betweenelements other than parent/child/sibling.
AXPropertyName = str

# AXNode: A node in the accessibility tree.
class AXNode(ChromeTypeBase):
    def __init__(self,
                 nodeId: Union['AXNodeId'],
                 ignored: Union['bool'],
                 ignoredReasons: Optional['[AXProperty]'] = None,
                 role: Optional['AXValue'] = None,
                 name: Optional['AXValue'] = None,
                 description: Optional['AXValue'] = None,
                 value: Optional['AXValue'] = None,
                 properties: Optional['[AXProperty]'] = None,
                 childIds: Optional['[AXNodeId]'] = None,
                 backendDOMNodeId: Optional['DOM.BackendNodeId'] = None,
                 ):

        self.nodeId = nodeId
        self.ignored = ignored
        self.ignoredReasons = ignoredReasons
        self.role = role
        self.name = name
        self.description = description
        self.value = value
        self.properties = properties
        self.childIds = childIds
        self.backendDOMNodeId = backendDOMNodeId


class Accessibility(PayloadMixin):
    """ 
    """
    @classmethod
    def disable(cls):
        """Disables the accessibility domain.
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def enable(cls):
        """Enables the accessibility domain which causes `AXNodeId`s to remain consistent between method calls.
This turns on accessibility for the page, which can impact performance until accessibility is disabled.
        """
        return (
            cls.build_send_payload("enable", {
            }),
            None
        )

    @classmethod
    def getPartialAXTree(cls,
                         nodeId: Optional['DOM.NodeId'] = None,
                         backendNodeId: Optional['DOM.BackendNodeId'] = None,
                         objectId: Optional['Runtime.RemoteObjectId'] = None,
                         fetchRelatives: Optional['bool'] = None,
                         ):
        """Fetches the accessibility node and partial accessibility tree for this DOM node, if it exists.
        :param nodeId: Identifier of the node to get the partial accessibility tree for.
        :type nodeId: DOM.NodeId
        :param backendNodeId: Identifier of the backend node to get the partial accessibility tree for.
        :type backendNodeId: DOM.BackendNodeId
        :param objectId: JavaScript object id of the node wrapper to get the partial accessibility tree for.
        :type objectId: Runtime.RemoteObjectId
        :param fetchRelatives: Whether to fetch this nodes ancestors, siblings and children. Defaults to true.
        :type fetchRelatives: bool
        """
        return (
            cls.build_send_payload("getPartialAXTree", {
                "nodeId": nodeId,
                "backendNodeId": backendNodeId,
                "objectId": objectId,
                "fetchRelatives": fetchRelatives,
            }),
            cls.convert_payload({
                "nodes": {
                    "class": [AXNode],
                    "optional": False
                },
            })
        )

    @classmethod
    def getFullAXTree(cls):
        """Fetches the entire accessibility tree
        """
        return (
            cls.build_send_payload("getFullAXTree", {
            }),
            cls.convert_payload({
                "nodes": {
                    "class": [AXNode],
                    "optional": False
                },
            })
        )




# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)

# TouchPoint: 
class TouchPoint(ChromeTypeBase):
    def __init__(self,
                 x: Union['float'],
                 y: Union['float'],
                 radiusX: Optional['float'] = None,
                 radiusY: Optional['float'] = None,
                 rotationAngle: Optional['float'] = None,
                 force: Optional['float'] = None,
                 id: Optional['float'] = None,
                 ):

        self.x = x
        self.y = y
        self.radiusX = radiusX
        self.radiusY = radiusY
        self.rotationAngle = rotationAngle
        self.force = force
        self.id = id


# GestureSourceType: 
GestureSourceType = str

# TimeSinceEpoch: UTC time in seconds, counted from January 1, 1970.
TimeSinceEpoch = float

class Input(PayloadMixin):
    """ 
    """
    @classmethod
    def dispatchKeyEvent(cls,
                         type: Union['str'],
                         modifiers: Optional['int'] = None,
                         timestamp: Optional['TimeSinceEpoch'] = None,
                         text: Optional['str'] = None,
                         unmodifiedText: Optional['str'] = None,
                         keyIdentifier: Optional['str'] = None,
                         code: Optional['str'] = None,
                         key: Optional['str'] = None,
                         windowsVirtualKeyCode: Optional['int'] = None,
                         nativeVirtualKeyCode: Optional['int'] = None,
                         autoRepeat: Optional['bool'] = None,
                         isKeypad: Optional['bool'] = None,
                         isSystemKey: Optional['bool'] = None,
                         location: Optional['int'] = None,
                         ):
        """Dispatches a key event to the page.
        :param type: Type of the key event.
        :type type: str
        :param modifiers: Bit field representing pressed modifier keys. Alt=1, Ctrl=2, Meta/Command=4, Shift=8
(default: 0).
        :type modifiers: int
        :param timestamp: Time at which the event occurred.
        :type timestamp: TimeSinceEpoch
        :param text: Text as generated by processing a virtual key code with a keyboard layout. Not needed for
for `keyUp` and `rawKeyDown` events (default: "")
        :type text: str
        :param unmodifiedText: Text that would have been generated by the keyboard if no modifiers were pressed (except for
shift). Useful for shortcut (accelerator) key handling (default: "").
        :type unmodifiedText: str
        :param keyIdentifier: Unique key identifier (e.g., 'U+0041') (default: "").
        :type keyIdentifier: str
        :param code: Unique DOM defined string value for each physical key (e.g., 'KeyA') (default: "").
        :type code: str
        :param key: Unique DOM defined string value describing the meaning of the key in the context of active
modifiers, keyboard layout, etc (e.g., 'AltGr') (default: "").
        :type key: str
        :param windowsVirtualKeyCode: Windows virtual key code (default: 0).
        :type windowsVirtualKeyCode: int
        :param nativeVirtualKeyCode: Native virtual key code (default: 0).
        :type nativeVirtualKeyCode: int
        :param autoRepeat: Whether the event was generated from auto repeat (default: false).
        :type autoRepeat: bool
        :param isKeypad: Whether the event was generated from the keypad (default: false).
        :type isKeypad: bool
        :param isSystemKey: Whether the event was a system key event (default: false).
        :type isSystemKey: bool
        :param location: Whether the event was from the left or right side of the keyboard. 1=Left, 2=Right (default:
0).
        :type location: int
        """
        return (
            cls.build_send_payload("dispatchKeyEvent", {
                "type": type,
                "modifiers": modifiers,
                "timestamp": timestamp,
                "text": text,
                "unmodifiedText": unmodifiedText,
                "keyIdentifier": keyIdentifier,
                "code": code,
                "key": key,
                "windowsVirtualKeyCode": windowsVirtualKeyCode,
                "nativeVirtualKeyCode": nativeVirtualKeyCode,
                "autoRepeat": autoRepeat,
                "isKeypad": isKeypad,
                "isSystemKey": isSystemKey,
                "location": location,
            }),
            None
        )

    @classmethod
    def insertText(cls,
                   text: Union['str'],
                   ):
        """This method emulates inserting text that doesn't come from a key press,
for example an emoji keyboard or an IME.
        :param text: The text to insert.
        :type text: str
        """
        return (
            cls.build_send_payload("insertText", {
                "text": text,
            }),
            None
        )

    @classmethod
    def dispatchMouseEvent(cls,
                           type: Union['str'],
                           x: Union['float'],
                           y: Union['float'],
                           modifiers: Optional['int'] = None,
                           timestamp: Optional['TimeSinceEpoch'] = None,
                           button: Optional['str'] = None,
                           clickCount: Optional['int'] = None,
                           deltaX: Optional['float'] = None,
                           deltaY: Optional['float'] = None,
                           ):
        """Dispatches a mouse event to the page.
        :param type: Type of the mouse event.
        :type type: str
        :param x: X coordinate of the event relative to the main frame's viewport in CSS pixels.
        :type x: float
        :param y: Y coordinate of the event relative to the main frame's viewport in CSS pixels. 0 refers to
the top of the viewport and Y increases as it proceeds towards the bottom of the viewport.
        :type y: float
        :param modifiers: Bit field representing pressed modifier keys. Alt=1, Ctrl=2, Meta/Command=4, Shift=8
(default: 0).
        :type modifiers: int
        :param timestamp: Time at which the event occurred.
        :type timestamp: TimeSinceEpoch
        :param button: Mouse button (default: "none").
        :type button: str
        :param clickCount: Number of times the mouse button was clicked (default: 0).
        :type clickCount: int
        :param deltaX: X delta in CSS pixels for mouse wheel event (default: 0).
        :type deltaX: float
        :param deltaY: Y delta in CSS pixels for mouse wheel event (default: 0).
        :type deltaY: float
        """
        return (
            cls.build_send_payload("dispatchMouseEvent", {
                "type": type,
                "x": x,
                "y": y,
                "modifiers": modifiers,
                "timestamp": timestamp,
                "button": button,
                "clickCount": clickCount,
                "deltaX": deltaX,
                "deltaY": deltaY,
            }),
            None
        )

    @classmethod
    def dispatchTouchEvent(cls,
                           type: Union['str'],
                           touchPoints: Union['[TouchPoint]'],
                           modifiers: Optional['int'] = None,
                           timestamp: Optional['TimeSinceEpoch'] = None,
                           ):
        """Dispatches a touch event to the page.
        :param type: Type of the touch event. TouchEnd and TouchCancel must not contain any touch points, while
TouchStart and TouchMove must contains at least one.
        :type type: str
        :param touchPoints: Active touch points on the touch device. One event per any changed point (compared to
previous touch event in a sequence) is generated, emulating pressing/moving/releasing points
one by one.
        :type touchPoints: [TouchPoint]
        :param modifiers: Bit field representing pressed modifier keys. Alt=1, Ctrl=2, Meta/Command=4, Shift=8
(default: 0).
        :type modifiers: int
        :param timestamp: Time at which the event occurred.
        :type timestamp: TimeSinceEpoch
        """
        return (
            cls.build_send_payload("dispatchTouchEvent", {
                "type": type,
                "touchPoints": touchPoints,
                "modifiers": modifiers,
                "timestamp": timestamp,
            }),
            None
        )

    @classmethod
    def emulateTouchFromMouseEvent(cls,
                                   type: Union['str'],
                                   x: Union['int'],
                                   y: Union['int'],
                                   button: Union['str'],
                                   timestamp: Optional['TimeSinceEpoch'] = None,
                                   deltaX: Optional['float'] = None,
                                   deltaY: Optional['float'] = None,
                                   modifiers: Optional['int'] = None,
                                   clickCount: Optional['int'] = None,
                                   ):
        """Emulates touch event from the mouse event parameters.
        :param type: Type of the mouse event.
        :type type: str
        :param x: X coordinate of the mouse pointer in DIP.
        :type x: int
        :param y: Y coordinate of the mouse pointer in DIP.
        :type y: int
        :param button: Mouse button.
        :type button: str
        :param timestamp: Time at which the event occurred (default: current time).
        :type timestamp: TimeSinceEpoch
        :param deltaX: X delta in DIP for mouse wheel event (default: 0).
        :type deltaX: float
        :param deltaY: Y delta in DIP for mouse wheel event (default: 0).
        :type deltaY: float
        :param modifiers: Bit field representing pressed modifier keys. Alt=1, Ctrl=2, Meta/Command=4, Shift=8
(default: 0).
        :type modifiers: int
        :param clickCount: Number of times the mouse button was clicked (default: 0).
        :type clickCount: int
        """
        return (
            cls.build_send_payload("emulateTouchFromMouseEvent", {
                "type": type,
                "x": x,
                "y": y,
                "button": button,
                "timestamp": timestamp,
                "deltaX": deltaX,
                "deltaY": deltaY,
                "modifiers": modifiers,
                "clickCount": clickCount,
            }),
            None
        )

    @classmethod
    def setIgnoreInputEvents(cls,
                             ignore: Union['bool'],
                             ):
        """Ignores input events (useful while auditing page).
        :param ignore: Ignores input events processing when set to true.
        :type ignore: bool
        """
        return (
            cls.build_send_payload("setIgnoreInputEvents", {
                "ignore": ignore,
            }),
            None
        )

    @classmethod
    def synthesizePinchGesture(cls,
                               x: Union['float'],
                               y: Union['float'],
                               scaleFactor: Union['float'],
                               relativeSpeed: Optional['int'] = None,
                               gestureSourceType: Optional['GestureSourceType'] = None,
                               ):
        """Synthesizes a pinch gesture over a time period by issuing appropriate touch events.
        :param x: X coordinate of the start of the gesture in CSS pixels.
        :type x: float
        :param y: Y coordinate of the start of the gesture in CSS pixels.
        :type y: float
        :param scaleFactor: Relative scale factor after zooming (>1.0 zooms in, <1.0 zooms out).
        :type scaleFactor: float
        :param relativeSpeed: Relative pointer speed in pixels per second (default: 800).
        :type relativeSpeed: int
        :param gestureSourceType: Which type of input events to be generated (default: 'default', which queries the platform
for the preferred input type).
        :type gestureSourceType: GestureSourceType
        """
        return (
            cls.build_send_payload("synthesizePinchGesture", {
                "x": x,
                "y": y,
                "scaleFactor": scaleFactor,
                "relativeSpeed": relativeSpeed,
                "gestureSourceType": gestureSourceType,
            }),
            None
        )

    @classmethod
    def synthesizeScrollGesture(cls,
                                x: Union['float'],
                                y: Union['float'],
                                xDistance: Optional['float'] = None,
                                yDistance: Optional['float'] = None,
                                xOverscroll: Optional['float'] = None,
                                yOverscroll: Optional['float'] = None,
                                preventFling: Optional['bool'] = None,
                                speed: Optional['int'] = None,
                                gestureSourceType: Optional['GestureSourceType'] = None,
                                repeatCount: Optional['int'] = None,
                                repeatDelayMs: Optional['int'] = None,
                                interactionMarkerName: Optional['str'] = None,
                                ):
        """Synthesizes a scroll gesture over a time period by issuing appropriate touch events.
        :param x: X coordinate of the start of the gesture in CSS pixels.
        :type x: float
        :param y: Y coordinate of the start of the gesture in CSS pixels.
        :type y: float
        :param xDistance: The distance to scroll along the X axis (positive to scroll left).
        :type xDistance: float
        :param yDistance: The distance to scroll along the Y axis (positive to scroll up).
        :type yDistance: float
        :param xOverscroll: The number of additional pixels to scroll back along the X axis, in addition to the given
distance.
        :type xOverscroll: float
        :param yOverscroll: The number of additional pixels to scroll back along the Y axis, in addition to the given
distance.
        :type yOverscroll: float
        :param preventFling: Prevent fling (default: true).
        :type preventFling: bool
        :param speed: Swipe speed in pixels per second (default: 800).
        :type speed: int
        :param gestureSourceType: Which type of input events to be generated (default: 'default', which queries the platform
for the preferred input type).
        :type gestureSourceType: GestureSourceType
        :param repeatCount: The number of times to repeat the gesture (default: 0).
        :type repeatCount: int
        :param repeatDelayMs: The number of milliseconds delay between each repeat. (default: 250).
        :type repeatDelayMs: int
        :param interactionMarkerName: The name of the interaction markers to generate, if not empty (default: "").
        :type interactionMarkerName: str
        """
        return (
            cls.build_send_payload("synthesizeScrollGesture", {
                "x": x,
                "y": y,
                "xDistance": xDistance,
                "yDistance": yDistance,
                "xOverscroll": xOverscroll,
                "yOverscroll": yOverscroll,
                "preventFling": preventFling,
                "speed": speed,
                "gestureSourceType": gestureSourceType,
                "repeatCount": repeatCount,
                "repeatDelayMs": repeatDelayMs,
                "interactionMarkerName": interactionMarkerName,
            }),
            None
        )

    @classmethod
    def synthesizeTapGesture(cls,
                             x: Union['float'],
                             y: Union['float'],
                             duration: Optional['int'] = None,
                             tapCount: Optional['int'] = None,
                             gestureSourceType: Optional['GestureSourceType'] = None,
                             ):
        """Synthesizes a tap gesture over a time period by issuing appropriate touch events.
        :param x: X coordinate of the start of the gesture in CSS pixels.
        :type x: float
        :param y: Y coordinate of the start of the gesture in CSS pixels.
        :type y: float
        :param duration: Duration between touchdown and touchup events in ms (default: 50).
        :type duration: int
        :param tapCount: Number of times to perform the tap (e.g. 2 for double tap, default: 1).
        :type tapCount: int
        :param gestureSourceType: Which type of input events to be generated (default: 'default', which queries the platform
for the preferred input type).
        :type gestureSourceType: GestureSourceType
        """
        return (
            cls.build_send_payload("synthesizeTapGesture", {
                "x": x,
                "y": y,
                "duration": duration,
                "tapCount": tapCount,
                "gestureSourceType": gestureSourceType,
            }),
            None
        )




# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)
from chromewhip.protocol import dom as DOM
from chromewhip.protocol import page as Page
from chromewhip.protocol import runtime as Runtime

# HighlightConfig: Configuration data for the highlighting of page elements.
class HighlightConfig(ChromeTypeBase):
    def __init__(self,
                 showInfo: Optional['bool'] = None,
                 showRulers: Optional['bool'] = None,
                 showExtensionLines: Optional['bool'] = None,
                 displayAsMaterial: Optional['bool'] = None,
                 contentColor: Optional['DOM.RGBA'] = None,
                 paddingColor: Optional['DOM.RGBA'] = None,
                 borderColor: Optional['DOM.RGBA'] = None,
                 marginColor: Optional['DOM.RGBA'] = None,
                 eventTargetColor: Optional['DOM.RGBA'] = None,
                 shapeColor: Optional['DOM.RGBA'] = None,
                 shapeMarginColor: Optional['DOM.RGBA'] = None,
                 selectorList: Optional['str'] = None,
                 cssGridColor: Optional['DOM.RGBA'] = None,
                 ):

        self.showInfo = showInfo
        self.showRulers = showRulers
        self.showExtensionLines = showExtensionLines
        self.displayAsMaterial = displayAsMaterial
        self.contentColor = contentColor
        self.paddingColor = paddingColor
        self.borderColor = borderColor
        self.marginColor = marginColor
        self.eventTargetColor = eventTargetColor
        self.shapeColor = shapeColor
        self.shapeMarginColor = shapeMarginColor
        self.selectorList = selectorList
        self.cssGridColor = cssGridColor


# InspectMode: 
InspectMode = str

class Overlay(PayloadMixin):
    """ This domain provides various functionality related to drawing atop the inspected page.
    """
    @classmethod
    def disable(cls):
        """Disables domain notifications.
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def enable(cls):
        """Enables domain notifications.
        """
        return (
            cls.build_send_payload("enable", {
            }),
            None
        )

    @classmethod
    def getHighlightObjectForTest(cls,
                                  nodeId: Union['DOM.NodeId'],
                                  ):
        """For testing.
        :param nodeId: Id of the node to get highlight object for.
        :type nodeId: DOM.NodeId
        """
        return (
            cls.build_send_payload("getHighlightObjectForTest", {
                "nodeId": nodeId,
            }),
            cls.convert_payload({
                "highlight": {
                    "class": dict,
                    "optional": False
                },
            })
        )

    @classmethod
    def hideHighlight(cls):
        """Hides any highlight.
        """
        return (
            cls.build_send_payload("hideHighlight", {
            }),
            None
        )

    @classmethod
    def highlightFrame(cls,
                       frameId: Union['Page.FrameId'],
                       contentColor: Optional['DOM.RGBA'] = None,
                       contentOutlineColor: Optional['DOM.RGBA'] = None,
                       ):
        """Highlights owner element of the frame with given id.
        :param frameId: Identifier of the frame to highlight.
        :type frameId: Page.FrameId
        :param contentColor: The content box highlight fill color (default: transparent).
        :type contentColor: DOM.RGBA
        :param contentOutlineColor: The content box highlight outline color (default: transparent).
        :type contentOutlineColor: DOM.RGBA
        """
        return (
            cls.build_send_payload("highlightFrame", {
                "frameId": frameId,
                "contentColor": contentColor,
                "contentOutlineColor": contentOutlineColor,
            }),
            None
        )

    @classmethod
    def highlightNode(cls,
                      highlightConfig: Union['HighlightConfig'],
                      nodeId: Optional['DOM.NodeId'] = None,
                      backendNodeId: Optional['DOM.BackendNodeId'] = None,
                      objectId: Optional['Runtime.RemoteObjectId'] = None,
                      ):
        """Highlights DOM node with given id or with the given JavaScript object wrapper. Either nodeId or
objectId must be specified.
        :param highlightConfig: A descriptor for the highlight appearance.
        :type highlightConfig: HighlightConfig
        :param nodeId: Identifier of the node to highlight.
        :type nodeId: DOM.NodeId
        :param backendNodeId: Identifier of the backend node to highlight.
        :type backendNodeId: DOM.BackendNodeId
        :param objectId: JavaScript object id of the node to be highlighted.
        :type objectId: Runtime.RemoteObjectId
        """
        return (
            cls.build_send_payload("highlightNode", {
                "highlightConfig": highlightConfig,
                "nodeId": nodeId,
                "backendNodeId": backendNodeId,
                "objectId": objectId,
            }),
            None
        )

    @classmethod
    def highlightQuad(cls,
                      quad: Union['DOM.Quad'],
                      color: Optional['DOM.RGBA'] = None,
                      outlineColor: Optional['DOM.RGBA'] = None,
                      ):
        """Highlights given quad. Coordinates are absolute with respect to the main frame viewport.
        :param quad: Quad to highlight
        :type quad: DOM.Quad
        :param color: The highlight fill color (default: transparent).
        :type color: DOM.RGBA
        :param outlineColor: The highlight outline color (default: transparent).
        :type outlineColor: DOM.RGBA
        """
        return (
            cls.build_send_payload("highlightQuad", {
                "quad": quad,
                "color": color,
                "outlineColor": outlineColor,
            }),
            None
        )

    @classmethod
    def highlightRect(cls,
                      x: Union['int'],
                      y: Union['int'],
                      width: Union['int'],
                      height: Union['int'],
                      color: Optional['DOM.RGBA'] = None,
                      outlineColor: Optional['DOM.RGBA'] = None,
                      ):
        """Highlights given rectangle. Coordinates are absolute with respect to the main frame viewport.
        :param x: X coordinate
        :type x: int
        :param y: Y coordinate
        :type y: int
        :param width: Rectangle width
        :type width: int
        :param height: Rectangle height
        :type height: int
        :param color: The highlight fill color (default: transparent).
        :type color: DOM.RGBA
        :param outlineColor: The highlight outline color (default: transparent).
        :type outlineColor: DOM.RGBA
        """
        return (
            cls.build_send_payload("highlightRect", {
                "x": x,
                "y": y,
                "width": width,
                "height": height,
                "color": color,
                "outlineColor": outlineColor,
            }),
            None
        )

    @classmethod
    def setInspectMode(cls,
                       mode: Union['InspectMode'],
                       highlightConfig: Optional['HighlightConfig'] = None,
                       ):
        """Enters the 'inspect' mode. In this mode, elements that user is hovering over are highlighted.
Backend then generates 'inspectNodeRequested' event upon element selection.
        :param mode: Set an inspection mode.
        :type mode: InspectMode
        :param highlightConfig: A descriptor for the highlight appearance of hovered-over nodes. May be omitted if `enabled
== false`.
        :type highlightConfig: HighlightConfig
        """
        return (
            cls.build_send_payload("setInspectMode", {
                "mode": mode,
                "highlightConfig": highlightConfig,
            }),
            None
        )

    @classmethod
    def setPausedInDebuggerMessage(cls,
                                   message: Optional['str'] = None,
                                   ):
        """
        :param message: The message to display, also triggers resume and step over controls.
        :type message: str
        """
        return (
            cls.build_send_payload("setPausedInDebuggerMessage", {
                "message": message,
            }),
            None
        )

    @classmethod
    def setShowDebugBorders(cls,
                            show: Union['bool'],
                            ):
        """Requests that backend shows debug borders on layers
        :param show: True for showing debug borders
        :type show: bool
        """
        return (
            cls.build_send_payload("setShowDebugBorders", {
                "show": show,
            }),
            None
        )

    @classmethod
    def setShowFPSCounter(cls,
                          show: Union['bool'],
                          ):
        """Requests that backend shows the FPS counter
        :param show: True for showing the FPS counter
        :type show: bool
        """
        return (
            cls.build_send_payload("setShowFPSCounter", {
                "show": show,
            }),
            None
        )

    @classmethod
    def setShowPaintRects(cls,
                          result: Union['bool'],
                          ):
        """Requests that backend shows paint rectangles
        :param result: True for showing paint rectangles
        :type result: bool
        """
        return (
            cls.build_send_payload("setShowPaintRects", {
                "result": result,
            }),
            None
        )

    @classmethod
    def setShowScrollBottleneckRects(cls,
                                     show: Union['bool'],
                                     ):
        """Requests that backend shows scroll bottleneck rects
        :param show: True for showing scroll bottleneck rects
        :type show: bool
        """
        return (
            cls.build_send_payload("setShowScrollBottleneckRects", {
                "show": show,
            }),
            None
        )

    @classmethod
    def setShowHitTestBorders(cls,
                              show: Union['bool'],
                              ):
        """Requests that backend shows hit-test borders on layers
        :param show: True for showing hit-test borders
        :type show: bool
        """
        return (
            cls.build_send_payload("setShowHitTestBorders", {
                "show": show,
            }),
            None
        )

    @classmethod
    def setShowViewportSizeOnResize(cls,
                                    show: Union['bool'],
                                    ):
        """Paints viewport size upon main frame resize.
        :param show: Whether to paint size or not.
        :type show: bool
        """
        return (
            cls.build_send_payload("setShowViewportSizeOnResize", {
                "show": show,
            }),
            None
        )

    @classmethod
    def setSuspended(cls,
                     suspended: Union['bool'],
                     ):
        """
        :param suspended: Whether overlay should be suspended and not consume any resources until resumed.
        :type suspended: bool
        """
        return (
            cls.build_send_payload("setSuspended", {
                "suspended": suspended,
            }),
            None
        )



class InspectNodeRequestedEvent(BaseEvent):

    js_name = 'Overlay.inspectNodeRequested'
    hashable = ['backendNodeId']
    is_hashable = True

    def __init__(self,
                 backendNodeId: Union['DOM.BackendNodeId', dict],
                 ):
        if isinstance(backendNodeId, dict):
            backendNodeId = DOM.BackendNodeId(**backendNodeId)
        self.backendNodeId = backendNodeId

    @classmethod
    def build_hash(cls, backendNodeId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class NodeHighlightRequestedEvent(BaseEvent):

    js_name = 'Overlay.nodeHighlightRequested'
    hashable = ['nodeId']
    is_hashable = True

    def __init__(self,
                 nodeId: Union['DOM.NodeId', dict],
                 ):
        if isinstance(nodeId, dict):
            nodeId = DOM.NodeId(**nodeId)
        self.nodeId = nodeId

    @classmethod
    def build_hash(cls, nodeId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class ScreenshotRequestedEvent(BaseEvent):

    js_name = 'Overlay.screenshotRequested'
    hashable = []
    is_hashable = False

    def __init__(self,
                 viewport: Union['Page.Viewport', dict],
                 ):
        if isinstance(viewport, dict):
            viewport = Page.Viewport(**viewport)
        self.viewport = viewport

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')



import copy
import json
import logging
import re
import sys


class PayloadMixin:
    @classmethod
    def build_send_payload(cls, method: str, params: dict):
        return {
            "method": ".".join([cls.__name__, method]),
            "params": {k: v for k, v in params.items() if v is not None}
        }

    @classmethod
    def convert_payload(cls, types: dict):
        def convert(result: dict):
            """

            :param result:
            :return:
            """
            types_ = copy.copy(types)
            for name, val in result.items():
                try:
                    expected_ = types_.pop(name)
                    expected_type_ = expected_['class']
                except KeyError:
                    raise KeyError('name %s not in expected payload of %s' % (name, types))
                if issubclass(expected_type_, ChromeTypeBase):
                    result[name] = expected_type_(**val)
                elif re.match(r'.*Id$', name) and isinstance(val, str):
                    result[name] = expected_type_(val)
                elif not isinstance(val, expected_type_):
                    raise ValueError('%s is not expected type %s, instead is %s' % (val, expected_type_, val))
            for rn, rv in types_.items():
                if not rv.get('optional', False):
                    raise ValueError('expected payload param "%s" is missing!' % rn)
            return result
        return convert


log = logging.getLogger(__name__)


class BaseEvent:
    js_name = 'chromewhipBaseEvent'
    hashable = []
    is_hashable = False

    def hash_(self):
        hashable_params = {}
        for k, v in self.__dict__.items():
            if k in self.hashable:
                hashable_params[k] = v
            else:
                try:
                    hashable_params['%sId' % k] = v.id
                except KeyError:
                    pass
                except AttributeError:
                    # TODO: make better, fails for event that has 'timestamp` as a param
                    pass
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in hashable_params.items()])
        h = '{}:{}'.format(self.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


# TODO: how do
def json_to_event(payload) -> BaseEvent:
    try:
        prot_name, js_event = payload['method'].split('.')
    except KeyError:
        log.error('invalid event JSON, must have a "method" key')
        return None
    except ValueError:
        log.error('invalid method name "%s", must contain a module and event joined with a "."' % payload['method'])
        return None
    module_name = 'chromewhip.protocol.%s' % prot_name.lower()
    try:
        prot_module = sys.modules[module_name]
    except KeyError:
        msg = '"%s" is not available in sys.modules!' % module_name
        log.error(msg)
        raise KeyError(msg)
    py_event_name = '{}{}Event'.format(js_event[0].upper(), js_event[1:])
    event_cls = getattr(prot_module, py_event_name)
    try:
        result = event_cls(**payload['params'])
    except TypeError as e:
        raise TypeError('%s unable to deserialise: %s' % (event_cls.__name__, e))
    return result


class ChromeTypeBase:

    def to_dict(self):
        return self.__dict__


class ChromewhipJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, BaseEvent):
            return {'method': obj.js_name, 'params': obj.__dict__}

        if isinstance(obj, ChromeTypeBase):
            return obj.__dict__

        return json.JSONEncoder.default(self, obj)



# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)

# StorageType: Enum of possible storage types.
StorageType = str

# UsageForType: Usage for a storage type.
class UsageForType(ChromeTypeBase):
    def __init__(self,
                 storageType: Union['StorageType'],
                 usage: Union['float'],
                 ):

        self.storageType = storageType
        self.usage = usage


class Storage(PayloadMixin):
    """ 
    """
    @classmethod
    def clearDataForOrigin(cls,
                           origin: Union['str'],
                           storageTypes: Union['str'],
                           ):
        """Clears storage for origin.
        :param origin: Security origin.
        :type origin: str
        :param storageTypes: Comma separated origin names.
        :type storageTypes: str
        """
        return (
            cls.build_send_payload("clearDataForOrigin", {
                "origin": origin,
                "storageTypes": storageTypes,
            }),
            None
        )

    @classmethod
    def getUsageAndQuota(cls,
                         origin: Union['str'],
                         ):
        """Returns usage and quota in bytes.
        :param origin: Security origin.
        :type origin: str
        """
        return (
            cls.build_send_payload("getUsageAndQuota", {
                "origin": origin,
            }),
            cls.convert_payload({
                "usage": {
                    "class": float,
                    "optional": False
                },
                "quota": {
                    "class": float,
                    "optional": False
                },
                "usageBreakdown": {
                    "class": [UsageForType],
                    "optional": False
                },
            })
        )

    @classmethod
    def trackCacheStorageForOrigin(cls,
                                   origin: Union['str'],
                                   ):
        """Registers origin to be notified when an update occurs to its cache storage list.
        :param origin: Security origin.
        :type origin: str
        """
        return (
            cls.build_send_payload("trackCacheStorageForOrigin", {
                "origin": origin,
            }),
            None
        )

    @classmethod
    def trackIndexedDBForOrigin(cls,
                                origin: Union['str'],
                                ):
        """Registers origin to be notified when an update occurs to its IndexedDB.
        :param origin: Security origin.
        :type origin: str
        """
        return (
            cls.build_send_payload("trackIndexedDBForOrigin", {
                "origin": origin,
            }),
            None
        )

    @classmethod
    def untrackCacheStorageForOrigin(cls,
                                     origin: Union['str'],
                                     ):
        """Unregisters origin from receiving notifications for cache storage.
        :param origin: Security origin.
        :type origin: str
        """
        return (
            cls.build_send_payload("untrackCacheStorageForOrigin", {
                "origin": origin,
            }),
            None
        )

    @classmethod
    def untrackIndexedDBForOrigin(cls,
                                  origin: Union['str'],
                                  ):
        """Unregisters origin from receiving notifications for IndexedDB.
        :param origin: Security origin.
        :type origin: str
        """
        return (
            cls.build_send_payload("untrackIndexedDBForOrigin", {
                "origin": origin,
            }),
            None
        )



class CacheStorageContentUpdatedEvent(BaseEvent):

    js_name = 'Storage.cacheStorageContentUpdated'
    hashable = []
    is_hashable = False

    def __init__(self,
                 origin: Union['str', dict],
                 cacheName: Union['str', dict],
                 ):
        if isinstance(origin, dict):
            origin = str(**origin)
        self.origin = origin
        if isinstance(cacheName, dict):
            cacheName = str(**cacheName)
        self.cacheName = cacheName

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class CacheStorageListUpdatedEvent(BaseEvent):

    js_name = 'Storage.cacheStorageListUpdated'
    hashable = []
    is_hashable = False

    def __init__(self,
                 origin: Union['str', dict],
                 ):
        if isinstance(origin, dict):
            origin = str(**origin)
        self.origin = origin

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class IndexedDBContentUpdatedEvent(BaseEvent):

    js_name = 'Storage.indexedDBContentUpdated'
    hashable = []
    is_hashable = False

    def __init__(self,
                 origin: Union['str', dict],
                 databaseName: Union['str', dict],
                 objectStoreName: Union['str', dict],
                 ):
        if isinstance(origin, dict):
            origin = str(**origin)
        self.origin = origin
        if isinstance(databaseName, dict):
            databaseName = str(**databaseName)
        self.databaseName = databaseName
        if isinstance(objectStoreName, dict):
            objectStoreName = str(**objectStoreName)
        self.objectStoreName = objectStoreName

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class IndexedDBListUpdatedEvent(BaseEvent):

    js_name = 'Storage.indexedDBListUpdated'
    hashable = []
    is_hashable = False

    def __init__(self,
                 origin: Union['str', dict],
                 ):
        if isinstance(origin, dict):
            origin = str(**origin)
        self.origin = origin

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')



# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)

# ScriptId: Unique script identifier.
ScriptId = str

# RemoteObjectId: Unique object identifier.
RemoteObjectId = str

# UnserializableValue: Primitive value which cannot be JSON-stringified. Includes values `-0`, `NaN`, `Infinity`,`-Infinity`, and bigint literals.
UnserializableValue = str

# RemoteObject: Mirror object referencing original JavaScript object.
class RemoteObject(ChromeTypeBase):
    def __init__(self,
                 type: Union['str'],
                 subtype: Optional['str'] = None,
                 className: Optional['str'] = None,
                 value: Optional['Any'] = None,
                 unserializableValue: Optional['UnserializableValue'] = None,
                 description: Optional['str'] = None,
                 objectId: Optional['RemoteObjectId'] = None,
                 preview: Optional['ObjectPreview'] = None,
                 customPreview: Optional['CustomPreview'] = None,
                 ):

        self.type = type
        self.subtype = subtype
        self.className = className
        self.value = value
        self.unserializableValue = unserializableValue
        self.description = description
        self.objectId = objectId
        self.preview = preview
        self.customPreview = customPreview


# CustomPreview: 
class CustomPreview(ChromeTypeBase):
    def __init__(self,
                 header: Union['str'],
                 bodyGetterId: Optional['RemoteObjectId'] = None,
                 ):

        self.header = header
        self.bodyGetterId = bodyGetterId


# ObjectPreview: Object containing abbreviated remote object value.
class ObjectPreview(ChromeTypeBase):
    def __init__(self,
                 type: Union['str'],
                 overflow: Union['bool'],
                 properties: Union['[PropertyPreview]'],
                 subtype: Optional['str'] = None,
                 description: Optional['str'] = None,
                 entries: Optional['[EntryPreview]'] = None,
                 ):

        self.type = type
        self.subtype = subtype
        self.description = description
        self.overflow = overflow
        self.properties = properties
        self.entries = entries


# PropertyPreview: 
class PropertyPreview(ChromeTypeBase):
    def __init__(self,
                 name: Union['str'],
                 type: Union['str'],
                 value: Optional['str'] = None,
                 valuePreview: Optional['ObjectPreview'] = None,
                 subtype: Optional['str'] = None,
                 ):

        self.name = name
        self.type = type
        self.value = value
        self.valuePreview = valuePreview
        self.subtype = subtype


# EntryPreview: 
class EntryPreview(ChromeTypeBase):
    def __init__(self,
                 value: Union['ObjectPreview'],
                 key: Optional['ObjectPreview'] = None,
                 ):

        self.key = key
        self.value = value


# PropertyDescriptor: Object property descriptor.
class PropertyDescriptor(ChromeTypeBase):
    def __init__(self,
                 name: Union['str'],
                 configurable: Union['bool'],
                 enumerable: Union['bool'],
                 value: Optional['RemoteObject'] = None,
                 writable: Optional['bool'] = None,
                 get: Optional['RemoteObject'] = None,
                 set: Optional['RemoteObject'] = None,
                 wasThrown: Optional['bool'] = None,
                 isOwn: Optional['bool'] = None,
                 symbol: Optional['RemoteObject'] = None,
                 ):

        self.name = name
        self.value = value
        self.writable = writable
        self.get = get
        self.set = set
        self.configurable = configurable
        self.enumerable = enumerable
        self.wasThrown = wasThrown
        self.isOwn = isOwn
        self.symbol = symbol


# InternalPropertyDescriptor: Object internal property descriptor. This property isn't normally visible in JavaScript code.
class InternalPropertyDescriptor(ChromeTypeBase):
    def __init__(self,
                 name: Union['str'],
                 value: Optional['RemoteObject'] = None,
                 ):

        self.name = name
        self.value = value


# CallArgument: Represents function call argument. Either remote object id `objectId`, primitive `value`,unserializable primitive value or neither of (for undefined) them should be specified.
class CallArgument(ChromeTypeBase):
    def __init__(self,
                 value: Optional['Any'] = None,
                 unserializableValue: Optional['UnserializableValue'] = None,
                 objectId: Optional['RemoteObjectId'] = None,
                 ):

        self.value = value
        self.unserializableValue = unserializableValue
        self.objectId = objectId


# ExecutionContextId: Id of an execution context.
ExecutionContextId = int

# ExecutionContextDescription: Description of an isolated world.
class ExecutionContextDescription(ChromeTypeBase):
    def __init__(self,
                 id: Union['ExecutionContextId'],
                 origin: Union['str'],
                 name: Union['str'],
                 auxData: Optional['dict'] = None,
                 ):

        self.id = id
        self.origin = origin
        self.name = name
        self.auxData = auxData


# ExceptionDetails: Detailed information about exception (or error) that was thrown during script compilation orexecution.
class ExceptionDetails(ChromeTypeBase):
    def __init__(self,
                 exceptionId: Union['int'],
                 text: Union['str'],
                 lineNumber: Union['int'],
                 columnNumber: Union['int'],
                 scriptId: Optional['ScriptId'] = None,
                 url: Optional['str'] = None,
                 stackTrace: Optional['StackTrace'] = None,
                 exception: Optional['RemoteObject'] = None,
                 executionContextId: Optional['ExecutionContextId'] = None,
                 ):

        self.exceptionId = exceptionId
        self.text = text
        self.lineNumber = lineNumber
        self.columnNumber = columnNumber
        self.scriptId = scriptId
        self.url = url
        self.stackTrace = stackTrace
        self.exception = exception
        self.executionContextId = executionContextId


# Timestamp: Number of milliseconds since epoch.
Timestamp = float

# TimeDelta: Number of milliseconds.
TimeDelta = float

# CallFrame: Stack entry for runtime errors and assertions.
class CallFrame(ChromeTypeBase):
    def __init__(self,
                 functionName: Union['str'],
                 scriptId: Union['ScriptId'],
                 url: Union['str'],
                 lineNumber: Union['int'],
                 columnNumber: Union['int'],
                 ):

        self.functionName = functionName
        self.scriptId = scriptId
        self.url = url
        self.lineNumber = lineNumber
        self.columnNumber = columnNumber


# StackTrace: Call frames for assertions or error messages.
class StackTrace(ChromeTypeBase):
    def __init__(self,
                 callFrames: Union['[CallFrame]'],
                 description: Optional['str'] = None,
                 parent: Optional['StackTrace'] = None,
                 parentId: Optional['StackTraceId'] = None,
                 ):

        self.description = description
        self.callFrames = callFrames
        self.parent = parent
        self.parentId = parentId


# UniqueDebuggerId: Unique identifier of current debugger.
UniqueDebuggerId = str

# StackTraceId: If `debuggerId` is set stack trace comes from another debugger and can be resolved there. Thisallows to track cross-debugger calls. See `Runtime.StackTrace` and `Debugger.paused` for usages.
class StackTraceId(ChromeTypeBase):
    def __init__(self,
                 id: Union['str'],
                 debuggerId: Optional['UniqueDebuggerId'] = None,
                 ):

        self.id = id
        self.debuggerId = debuggerId


class Runtime(PayloadMixin):
    """ Runtime domain exposes JavaScript runtime by means of remote evaluation and mirror objects.
Evaluation results are returned as mirror object that expose object type, string representation
and unique identifier that can be used for further object reference. Original objects are
maintained in memory unless they are either explicitly released or are released along with the
other objects in their object group.
    """
    @classmethod
    def awaitPromise(cls,
                     promiseObjectId: Union['RemoteObjectId'],
                     returnByValue: Optional['bool'] = None,
                     generatePreview: Optional['bool'] = None,
                     ):
        """Add handler to promise with given promise object id.
        :param promiseObjectId: Identifier of the promise.
        :type promiseObjectId: RemoteObjectId
        :param returnByValue: Whether the result is expected to be a JSON object that should be sent by value.
        :type returnByValue: bool
        :param generatePreview: Whether preview should be generated for the result.
        :type generatePreview: bool
        """
        return (
            cls.build_send_payload("awaitPromise", {
                "promiseObjectId": promiseObjectId,
                "returnByValue": returnByValue,
                "generatePreview": generatePreview,
            }),
            cls.convert_payload({
                "result": {
                    "class": RemoteObject,
                    "optional": False
                },
                "exceptionDetails": {
                    "class": ExceptionDetails,
                    "optional": True
                },
            })
        )

    @classmethod
    def callFunctionOn(cls,
                       functionDeclaration: Union['str'],
                       objectId: Optional['RemoteObjectId'] = None,
                       arguments: Optional['[CallArgument]'] = None,
                       silent: Optional['bool'] = None,
                       returnByValue: Optional['bool'] = None,
                       generatePreview: Optional['bool'] = None,
                       userGesture: Optional['bool'] = None,
                       awaitPromise: Optional['bool'] = None,
                       executionContextId: Optional['ExecutionContextId'] = None,
                       objectGroup: Optional['str'] = None,
                       ):
        """Calls function with given declaration on the given object. Object group of the result is
inherited from the target object.
        :param functionDeclaration: Declaration of the function to call.
        :type functionDeclaration: str
        :param objectId: Identifier of the object to call function on. Either objectId or executionContextId should
be specified.
        :type objectId: RemoteObjectId
        :param arguments: Call arguments. All call arguments must belong to the same JavaScript world as the target
object.
        :type arguments: [CallArgument]
        :param silent: In silent mode exceptions thrown during evaluation are not reported and do not pause
execution. Overrides `setPauseOnException` state.
        :type silent: bool
        :param returnByValue: Whether the result is expected to be a JSON object which should be sent by value.
        :type returnByValue: bool
        :param generatePreview: Whether preview should be generated for the result.
        :type generatePreview: bool
        :param userGesture: Whether execution should be treated as initiated by user in the UI.
        :type userGesture: bool
        :param awaitPromise: Whether execution should `await` for resulting value and return once awaited promise is
resolved.
        :type awaitPromise: bool
        :param executionContextId: Specifies execution context which global object will be used to call function on. Either
executionContextId or objectId should be specified.
        :type executionContextId: ExecutionContextId
        :param objectGroup: Symbolic group name that can be used to release multiple objects. If objectGroup is not
specified and objectId is, objectGroup will be inherited from object.
        :type objectGroup: str
        """
        return (
            cls.build_send_payload("callFunctionOn", {
                "functionDeclaration": functionDeclaration,
                "objectId": objectId,
                "arguments": arguments,
                "silent": silent,
                "returnByValue": returnByValue,
                "generatePreview": generatePreview,
                "userGesture": userGesture,
                "awaitPromise": awaitPromise,
                "executionContextId": executionContextId,
                "objectGroup": objectGroup,
            }),
            cls.convert_payload({
                "result": {
                    "class": RemoteObject,
                    "optional": False
                },
                "exceptionDetails": {
                    "class": ExceptionDetails,
                    "optional": True
                },
            })
        )

    @classmethod
    def compileScript(cls,
                      expression: Union['str'],
                      sourceURL: Union['str'],
                      persistScript: Union['bool'],
                      executionContextId: Optional['ExecutionContextId'] = None,
                      ):
        """Compiles expression.
        :param expression: Expression to compile.
        :type expression: str
        :param sourceURL: Source url to be set for the script.
        :type sourceURL: str
        :param persistScript: Specifies whether the compiled script should be persisted.
        :type persistScript: bool
        :param executionContextId: Specifies in which execution context to perform script run. If the parameter is omitted the
evaluation will be performed in the context of the inspected page.
        :type executionContextId: ExecutionContextId
        """
        return (
            cls.build_send_payload("compileScript", {
                "expression": expression,
                "sourceURL": sourceURL,
                "persistScript": persistScript,
                "executionContextId": executionContextId,
            }),
            cls.convert_payload({
                "scriptId": {
                    "class": ScriptId,
                    "optional": True
                },
                "exceptionDetails": {
                    "class": ExceptionDetails,
                    "optional": True
                },
            })
        )

    @classmethod
    def disable(cls):
        """Disables reporting of execution contexts creation.
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def discardConsoleEntries(cls):
        """Discards collected exceptions and console API calls.
        """
        return (
            cls.build_send_payload("discardConsoleEntries", {
            }),
            None
        )

    @classmethod
    def enable(cls):
        """Enables reporting of execution contexts creation by means of `executionContextCreated` event.
When the reporting gets enabled the event will be sent immediately for each existing execution
context.
        """
        return (
            cls.build_send_payload("enable", {
            }),
            None
        )

    @classmethod
    def evaluate(cls,
                 expression: Union['str'],
                 objectGroup: Optional['str'] = None,
                 includeCommandLineAPI: Optional['bool'] = None,
                 silent: Optional['bool'] = None,
                 contextId: Optional['ExecutionContextId'] = None,
                 returnByValue: Optional['bool'] = None,
                 generatePreview: Optional['bool'] = None,
                 userGesture: Optional['bool'] = None,
                 awaitPromise: Optional['bool'] = None,
                 throwOnSideEffect: Optional['bool'] = None,
                 timeout: Optional['TimeDelta'] = None,
                 ):
        """Evaluates expression on global object.
        :param expression: Expression to evaluate.
        :type expression: str
        :param objectGroup: Symbolic group name that can be used to release multiple objects.
        :type objectGroup: str
        :param includeCommandLineAPI: Determines whether Command Line API should be available during the evaluation.
        :type includeCommandLineAPI: bool
        :param silent: In silent mode exceptions thrown during evaluation are not reported and do not pause
execution. Overrides `setPauseOnException` state.
        :type silent: bool
        :param contextId: Specifies in which execution context to perform evaluation. If the parameter is omitted the
evaluation will be performed in the context of the inspected page.
        :type contextId: ExecutionContextId
        :param returnByValue: Whether the result is expected to be a JSON object that should be sent by value.
        :type returnByValue: bool
        :param generatePreview: Whether preview should be generated for the result.
        :type generatePreview: bool
        :param userGesture: Whether execution should be treated as initiated by user in the UI.
        :type userGesture: bool
        :param awaitPromise: Whether execution should `await` for resulting value and return once awaited promise is
resolved.
        :type awaitPromise: bool
        :param throwOnSideEffect: Whether to throw an exception if side effect cannot be ruled out during evaluation.
        :type throwOnSideEffect: bool
        :param timeout: Terminate execution after timing out (number of milliseconds).
        :type timeout: TimeDelta
        """
        return (
            cls.build_send_payload("evaluate", {
                "expression": expression,
                "objectGroup": objectGroup,
                "includeCommandLineAPI": includeCommandLineAPI,
                "silent": silent,
                "contextId": contextId,
                "returnByValue": returnByValue,
                "generatePreview": generatePreview,
                "userGesture": userGesture,
                "awaitPromise": awaitPromise,
                "throwOnSideEffect": throwOnSideEffect,
                "timeout": timeout,
            }),
            cls.convert_payload({
                "result": {
                    "class": RemoteObject,
                    "optional": False
                },
                "exceptionDetails": {
                    "class": ExceptionDetails,
                    "optional": True
                },
            })
        )

    @classmethod
    def getIsolateId(cls):
        """Returns the isolate id.
        """
        return (
            cls.build_send_payload("getIsolateId", {
            }),
            cls.convert_payload({
                "id": {
                    "class": str,
                    "optional": False
                },
            })
        )

    @classmethod
    def getHeapUsage(cls):
        """Returns the JavaScript heap usage.
It is the total usage of the corresponding isolate not scoped to a particular Runtime.
        """
        return (
            cls.build_send_payload("getHeapUsage", {
            }),
            cls.convert_payload({
                "usedSize": {
                    "class": float,
                    "optional": False
                },
                "totalSize": {
                    "class": float,
                    "optional": False
                },
            })
        )

    @classmethod
    def getProperties(cls,
                      objectId: Union['RemoteObjectId'],
                      ownProperties: Optional['bool'] = None,
                      accessorPropertiesOnly: Optional['bool'] = None,
                      generatePreview: Optional['bool'] = None,
                      ):
        """Returns properties of a given object. Object group of the result is inherited from the target
object.
        :param objectId: Identifier of the object to return properties for.
        :type objectId: RemoteObjectId
        :param ownProperties: If true, returns properties belonging only to the element itself, not to its prototype
chain.
        :type ownProperties: bool
        :param accessorPropertiesOnly: If true, returns accessor properties (with getter/setter) only; internal properties are not
returned either.
        :type accessorPropertiesOnly: bool
        :param generatePreview: Whether preview should be generated for the results.
        :type generatePreview: bool
        """
        return (
            cls.build_send_payload("getProperties", {
                "objectId": objectId,
                "ownProperties": ownProperties,
                "accessorPropertiesOnly": accessorPropertiesOnly,
                "generatePreview": generatePreview,
            }),
            cls.convert_payload({
                "result": {
                    "class": [PropertyDescriptor],
                    "optional": False
                },
                "internalProperties": {
                    "class": [InternalPropertyDescriptor],
                    "optional": True
                },
                "exceptionDetails": {
                    "class": ExceptionDetails,
                    "optional": True
                },
            })
        )

    @classmethod
    def globalLexicalScopeNames(cls,
                                executionContextId: Optional['ExecutionContextId'] = None,
                                ):
        """Returns all let, const and class variables from global scope.
        :param executionContextId: Specifies in which execution context to lookup global scope variables.
        :type executionContextId: ExecutionContextId
        """
        return (
            cls.build_send_payload("globalLexicalScopeNames", {
                "executionContextId": executionContextId,
            }),
            cls.convert_payload({
                "names": {
                    "class": [],
                    "optional": False
                },
            })
        )

    @classmethod
    def queryObjects(cls,
                     prototypeObjectId: Union['RemoteObjectId'],
                     objectGroup: Optional['str'] = None,
                     ):
        """
        :param prototypeObjectId: Identifier of the prototype to return objects for.
        :type prototypeObjectId: RemoteObjectId
        :param objectGroup: Symbolic group name that can be used to release the results.
        :type objectGroup: str
        """
        return (
            cls.build_send_payload("queryObjects", {
                "prototypeObjectId": prototypeObjectId,
                "objectGroup": objectGroup,
            }),
            cls.convert_payload({
                "objects": {
                    "class": RemoteObject,
                    "optional": False
                },
            })
        )

    @classmethod
    def releaseObject(cls,
                      objectId: Union['RemoteObjectId'],
                      ):
        """Releases remote object with given id.
        :param objectId: Identifier of the object to release.
        :type objectId: RemoteObjectId
        """
        return (
            cls.build_send_payload("releaseObject", {
                "objectId": objectId,
            }),
            None
        )

    @classmethod
    def releaseObjectGroup(cls,
                           objectGroup: Union['str'],
                           ):
        """Releases all remote objects that belong to a given group.
        :param objectGroup: Symbolic object group name.
        :type objectGroup: str
        """
        return (
            cls.build_send_payload("releaseObjectGroup", {
                "objectGroup": objectGroup,
            }),
            None
        )

    @classmethod
    def runIfWaitingForDebugger(cls):
        """Tells inspected instance to run if it was waiting for debugger to attach.
        """
        return (
            cls.build_send_payload("runIfWaitingForDebugger", {
            }),
            None
        )

    @classmethod
    def runScript(cls,
                  scriptId: Union['ScriptId'],
                  executionContextId: Optional['ExecutionContextId'] = None,
                  objectGroup: Optional['str'] = None,
                  silent: Optional['bool'] = None,
                  includeCommandLineAPI: Optional['bool'] = None,
                  returnByValue: Optional['bool'] = None,
                  generatePreview: Optional['bool'] = None,
                  awaitPromise: Optional['bool'] = None,
                  ):
        """Runs script with given id in a given context.
        :param scriptId: Id of the script to run.
        :type scriptId: ScriptId
        :param executionContextId: Specifies in which execution context to perform script run. If the parameter is omitted the
evaluation will be performed in the context of the inspected page.
        :type executionContextId: ExecutionContextId
        :param objectGroup: Symbolic group name that can be used to release multiple objects.
        :type objectGroup: str
        :param silent: In silent mode exceptions thrown during evaluation are not reported and do not pause
execution. Overrides `setPauseOnException` state.
        :type silent: bool
        :param includeCommandLineAPI: Determines whether Command Line API should be available during the evaluation.
        :type includeCommandLineAPI: bool
        :param returnByValue: Whether the result is expected to be a JSON object which should be sent by value.
        :type returnByValue: bool
        :param generatePreview: Whether preview should be generated for the result.
        :type generatePreview: bool
        :param awaitPromise: Whether execution should `await` for resulting value and return once awaited promise is
resolved.
        :type awaitPromise: bool
        """
        return (
            cls.build_send_payload("runScript", {
                "scriptId": scriptId,
                "executionContextId": executionContextId,
                "objectGroup": objectGroup,
                "silent": silent,
                "includeCommandLineAPI": includeCommandLineAPI,
                "returnByValue": returnByValue,
                "generatePreview": generatePreview,
                "awaitPromise": awaitPromise,
            }),
            cls.convert_payload({
                "result": {
                    "class": RemoteObject,
                    "optional": False
                },
                "exceptionDetails": {
                    "class": ExceptionDetails,
                    "optional": True
                },
            })
        )

    @classmethod
    def setAsyncCallStackDepth(cls,
                               maxDepth: Union['int'],
                               ):
        """Enables or disables async call stacks tracking.
        :param maxDepth: Maximum depth of async call stacks. Setting to `0` will effectively disable collecting async
call stacks (default).
        :type maxDepth: int
        """
        return (
            cls.build_send_payload("setAsyncCallStackDepth", {
                "maxDepth": maxDepth,
            }),
            None
        )

    @classmethod
    def setCustomObjectFormatterEnabled(cls,
                                        enabled: Union['bool'],
                                        ):
        """
        :param enabled: 
        :type enabled: bool
        """
        return (
            cls.build_send_payload("setCustomObjectFormatterEnabled", {
                "enabled": enabled,
            }),
            None
        )

    @classmethod
    def setMaxCallStackSizeToCapture(cls,
                                     size: Union['int'],
                                     ):
        """
        :param size: 
        :type size: int
        """
        return (
            cls.build_send_payload("setMaxCallStackSizeToCapture", {
                "size": size,
            }),
            None
        )

    @classmethod
    def terminateExecution(cls):
        """Terminate current or next JavaScript execution.
Will cancel the termination when the outer-most script execution ends.
        """
        return (
            cls.build_send_payload("terminateExecution", {
            }),
            None
        )

    @classmethod
    def addBinding(cls,
                   name: Union['str'],
                   executionContextId: Optional['ExecutionContextId'] = None,
                   ):
        """If executionContextId is empty, adds binding with the given name on the
global objects of all inspected contexts, including those created later,
bindings survive reloads.
If executionContextId is specified, adds binding only on global object of
given execution context.
Binding function takes exactly one argument, this argument should be string,
in case of any other input, function throws an exception.
Each binding function call produces Runtime.bindingCalled notification.
        :param name: 
        :type name: str
        :param executionContextId: 
        :type executionContextId: ExecutionContextId
        """
        return (
            cls.build_send_payload("addBinding", {
                "name": name,
                "executionContextId": executionContextId,
            }),
            None
        )

    @classmethod
    def removeBinding(cls,
                      name: Union['str'],
                      ):
        """This method does not remove binding function from global object but
unsubscribes current runtime agent from Runtime.bindingCalled notifications.
        :param name: 
        :type name: str
        """
        return (
            cls.build_send_payload("removeBinding", {
                "name": name,
            }),
            None
        )



class BindingCalledEvent(BaseEvent):

    js_name = 'Runtime.bindingCalled'
    hashable = ['executionContextId']
    is_hashable = True

    def __init__(self,
                 name: Union['str', dict],
                 payload: Union['str', dict],
                 executionContextId: Union['ExecutionContextId', dict],
                 ):
        if isinstance(name, dict):
            name = str(**name)
        self.name = name
        if isinstance(payload, dict):
            payload = str(**payload)
        self.payload = payload
        if isinstance(executionContextId, dict):
            executionContextId = ExecutionContextId(**executionContextId)
        self.executionContextId = executionContextId

    @classmethod
    def build_hash(cls, executionContextId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class ConsoleAPICalledEvent(BaseEvent):

    js_name = 'Runtime.consoleAPICalled'
    hashable = ['executionContextId']
    is_hashable = True

    def __init__(self,
                 type: Union['str', dict],
                 args: Union['[RemoteObject]', dict],
                 executionContextId: Union['ExecutionContextId', dict],
                 timestamp: Union['Timestamp', dict],
                 stackTrace: Union['StackTrace', dict, None] = None,
                 context: Union['str', dict, None] = None,
                 ):
        if isinstance(type, dict):
            type = str(**type)
        self.type = type
        if isinstance(args, dict):
            args = [RemoteObject](**args)
        self.args = args
        if isinstance(executionContextId, dict):
            executionContextId = ExecutionContextId(**executionContextId)
        self.executionContextId = executionContextId
        if isinstance(timestamp, dict):
            timestamp = Timestamp(**timestamp)
        self.timestamp = timestamp
        if isinstance(stackTrace, dict):
            stackTrace = StackTrace(**stackTrace)
        self.stackTrace = stackTrace
        if isinstance(context, dict):
            context = str(**context)
        self.context = context

    @classmethod
    def build_hash(cls, executionContextId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class ExceptionRevokedEvent(BaseEvent):

    js_name = 'Runtime.exceptionRevoked'
    hashable = ['exceptionId']
    is_hashable = True

    def __init__(self,
                 reason: Union['str', dict],
                 exceptionId: Union['int', dict],
                 ):
        if isinstance(reason, dict):
            reason = str(**reason)
        self.reason = reason
        if isinstance(exceptionId, dict):
            exceptionId = int(**exceptionId)
        self.exceptionId = exceptionId

    @classmethod
    def build_hash(cls, exceptionId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class ExceptionThrownEvent(BaseEvent):

    js_name = 'Runtime.exceptionThrown'
    hashable = []
    is_hashable = False

    def __init__(self,
                 timestamp: Union['Timestamp', dict],
                 exceptionDetails: Union['ExceptionDetails', dict],
                 ):
        if isinstance(timestamp, dict):
            timestamp = Timestamp(**timestamp)
        self.timestamp = timestamp
        if isinstance(exceptionDetails, dict):
            exceptionDetails = ExceptionDetails(**exceptionDetails)
        self.exceptionDetails = exceptionDetails

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class ExecutionContextCreatedEvent(BaseEvent):

    js_name = 'Runtime.executionContextCreated'
    hashable = ['contextId']
    is_hashable = True

    def __init__(self,
                 context: Union['ExecutionContextDescription', dict],
                 ):
        if isinstance(context, dict):
            context = ExecutionContextDescription(**context)
        self.context = context

    @classmethod
    def build_hash(cls, contextId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class ExecutionContextDestroyedEvent(BaseEvent):

    js_name = 'Runtime.executionContextDestroyed'
    hashable = ['executionContextId']
    is_hashable = True

    def __init__(self,
                 executionContextId: Union['ExecutionContextId', dict],
                 ):
        if isinstance(executionContextId, dict):
            executionContextId = ExecutionContextId(**executionContextId)
        self.executionContextId = executionContextId

    @classmethod
    def build_hash(cls, executionContextId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class ExecutionContextsClearedEvent(BaseEvent):

    js_name = 'Runtime.executionContextsCleared'
    hashable = []
    is_hashable = False

    def __init__(self):
        pass

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class InspectRequestedEvent(BaseEvent):

    js_name = 'Runtime.inspectRequested'
    hashable = []
    is_hashable = False

    def __init__(self,
                 object: Union['RemoteObject', dict],
                 hints: Union['dict', dict],
                 ):
        if isinstance(object, dict):
            object = RemoteObject(**object)
        self.object = object
        if isinstance(hints, dict):
            hints = dict(**hints)
        self.hints = hints

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')



import json
import traceback
import pprint

from aiohttp import web

from chromewhip.chrome import ChromewhipException


def json_error(message):
    # return web.Response(
    #     body=json.dumps({'error': message}).encode('utf-8'),
    #     content_type='application/json')
    # return web.Response(text=pprint.pformat({'error': message}))
    return web.Response(text=json.dumps({'error': message}, indent=4))

async def error_middleware(app, handler):
    async def middleware_handler(request):
        try:
            response = await handler(request)
            if response.status != 200:
                return json_error(response.message)
            return response
        except web.HTTPException as ex:
            return json_error(ex.reason)
        except ChromewhipException as ex:
            return json_error(ex.args[0])
        except Exception as ex:
            verbose_tb = traceback.format_exc()
            return json_error(verbose_tb)
    return middleware_handler



# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)

# Domain: Description of the protocol domain.
class Domain(ChromeTypeBase):
    def __init__(self,
                 name: Union['str'],
                 version: Union['str'],
                 ):

        self.name = name
        self.version = version


class Schema(PayloadMixin):
    """ This domain is deprecated.
    """
    @classmethod
    def getDomains(cls):
        """Returns supported domains.
        """
        return (
            cls.build_send_payload("getDomains", {
            }),
            cls.convert_payload({
                "domains": {
                    "class": [Domain],
                    "optional": False
                },
            })
        )




# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)
from chromewhip.protocol import dom as DOM

# LayerId: Unique Layer identifier.
LayerId = str

# SnapshotId: Unique snapshot identifier.
SnapshotId = str

# ScrollRect: Rectangle where scrolling happens on the main thread.
class ScrollRect(ChromeTypeBase):
    def __init__(self,
                 rect: Union['DOM.Rect'],
                 type: Union['str'],
                 ):

        self.rect = rect
        self.type = type


# StickyPositionConstraint: Sticky position constraints.
class StickyPositionConstraint(ChromeTypeBase):
    def __init__(self,
                 stickyBoxRect: Union['DOM.Rect'],
                 containingBlockRect: Union['DOM.Rect'],
                 nearestLayerShiftingStickyBox: Optional['LayerId'] = None,
                 nearestLayerShiftingContainingBlock: Optional['LayerId'] = None,
                 ):

        self.stickyBoxRect = stickyBoxRect
        self.containingBlockRect = containingBlockRect
        self.nearestLayerShiftingStickyBox = nearestLayerShiftingStickyBox
        self.nearestLayerShiftingContainingBlock = nearestLayerShiftingContainingBlock


# PictureTile: Serialized fragment of layer picture along with its offset within the layer.
class PictureTile(ChromeTypeBase):
    def __init__(self,
                 x: Union['float'],
                 y: Union['float'],
                 picture: Union['str'],
                 ):

        self.x = x
        self.y = y
        self.picture = picture


# Layer: Information about a compositing layer.
class Layer(ChromeTypeBase):
    def __init__(self,
                 layerId: Union['LayerId'],
                 offsetX: Union['float'],
                 offsetY: Union['float'],
                 width: Union['float'],
                 height: Union['float'],
                 paintCount: Union['int'],
                 drawsContent: Union['bool'],
                 parentLayerId: Optional['LayerId'] = None,
                 backendNodeId: Optional['DOM.BackendNodeId'] = None,
                 transform: Optional['[]'] = None,
                 anchorX: Optional['float'] = None,
                 anchorY: Optional['float'] = None,
                 anchorZ: Optional['float'] = None,
                 invisible: Optional['bool'] = None,
                 scrollRects: Optional['[ScrollRect]'] = None,
                 stickyPositionConstraint: Optional['StickyPositionConstraint'] = None,
                 ):

        self.layerId = layerId
        self.parentLayerId = parentLayerId
        self.backendNodeId = backendNodeId
        self.offsetX = offsetX
        self.offsetY = offsetY
        self.width = width
        self.height = height
        self.transform = transform
        self.anchorX = anchorX
        self.anchorY = anchorY
        self.anchorZ = anchorZ
        self.paintCount = paintCount
        self.drawsContent = drawsContent
        self.invisible = invisible
        self.scrollRects = scrollRects
        self.stickyPositionConstraint = stickyPositionConstraint


# PaintProfile: Array of timings, one per paint step.
PaintProfile = [float]

class LayerTree(PayloadMixin):
    """ 
    """
    @classmethod
    def compositingReasons(cls,
                           layerId: Union['LayerId'],
                           ):
        """Provides the reasons why the given layer was composited.
        :param layerId: The id of the layer for which we want to get the reasons it was composited.
        :type layerId: LayerId
        """
        return (
            cls.build_send_payload("compositingReasons", {
                "layerId": layerId,
            }),
            cls.convert_payload({
                "compositingReasons": {
                    "class": [],
                    "optional": False
                },
            })
        )

    @classmethod
    def disable(cls):
        """Disables compositing tree inspection.
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def enable(cls):
        """Enables compositing tree inspection.
        """
        return (
            cls.build_send_payload("enable", {
            }),
            None
        )

    @classmethod
    def loadSnapshot(cls,
                     tiles: Union['[PictureTile]'],
                     ):
        """Returns the snapshot identifier.
        :param tiles: An array of tiles composing the snapshot.
        :type tiles: [PictureTile]
        """
        return (
            cls.build_send_payload("loadSnapshot", {
                "tiles": tiles,
            }),
            cls.convert_payload({
                "snapshotId": {
                    "class": SnapshotId,
                    "optional": False
                },
            })
        )

    @classmethod
    def makeSnapshot(cls,
                     layerId: Union['LayerId'],
                     ):
        """Returns the layer snapshot identifier.
        :param layerId: The id of the layer.
        :type layerId: LayerId
        """
        return (
            cls.build_send_payload("makeSnapshot", {
                "layerId": layerId,
            }),
            cls.convert_payload({
                "snapshotId": {
                    "class": SnapshotId,
                    "optional": False
                },
            })
        )

    @classmethod
    def profileSnapshot(cls,
                        snapshotId: Union['SnapshotId'],
                        minRepeatCount: Optional['int'] = None,
                        minDuration: Optional['float'] = None,
                        clipRect: Optional['DOM.Rect'] = None,
                        ):
        """
        :param snapshotId: The id of the layer snapshot.
        :type snapshotId: SnapshotId
        :param minRepeatCount: The maximum number of times to replay the snapshot (1, if not specified).
        :type minRepeatCount: int
        :param minDuration: The minimum duration (in seconds) to replay the snapshot.
        :type minDuration: float
        :param clipRect: The clip rectangle to apply when replaying the snapshot.
        :type clipRect: DOM.Rect
        """
        return (
            cls.build_send_payload("profileSnapshot", {
                "snapshotId": snapshotId,
                "minRepeatCount": minRepeatCount,
                "minDuration": minDuration,
                "clipRect": clipRect,
            }),
            cls.convert_payload({
                "timings": {
                    "class": [PaintProfile],
                    "optional": False
                },
            })
        )

    @classmethod
    def releaseSnapshot(cls,
                        snapshotId: Union['SnapshotId'],
                        ):
        """Releases layer snapshot captured by the back-end.
        :param snapshotId: The id of the layer snapshot.
        :type snapshotId: SnapshotId
        """
        return (
            cls.build_send_payload("releaseSnapshot", {
                "snapshotId": snapshotId,
            }),
            None
        )

    @classmethod
    def replaySnapshot(cls,
                       snapshotId: Union['SnapshotId'],
                       fromStep: Optional['int'] = None,
                       toStep: Optional['int'] = None,
                       scale: Optional['float'] = None,
                       ):
        """Replays the layer snapshot and returns the resulting bitmap.
        :param snapshotId: The id of the layer snapshot.
        :type snapshotId: SnapshotId
        :param fromStep: The first step to replay from (replay from the very start if not specified).
        :type fromStep: int
        :param toStep: The last step to replay to (replay till the end if not specified).
        :type toStep: int
        :param scale: The scale to apply while replaying (defaults to 1).
        :type scale: float
        """
        return (
            cls.build_send_payload("replaySnapshot", {
                "snapshotId": snapshotId,
                "fromStep": fromStep,
                "toStep": toStep,
                "scale": scale,
            }),
            cls.convert_payload({
                "dataURL": {
                    "class": str,
                    "optional": False
                },
            })
        )

    @classmethod
    def snapshotCommandLog(cls,
                           snapshotId: Union['SnapshotId'],
                           ):
        """Replays the layer snapshot and returns canvas log.
        :param snapshotId: The id of the layer snapshot.
        :type snapshotId: SnapshotId
        """
        return (
            cls.build_send_payload("snapshotCommandLog", {
                "snapshotId": snapshotId,
            }),
            cls.convert_payload({
                "commandLog": {
                    "class": [],
                    "optional": False
                },
            })
        )



class LayerPaintedEvent(BaseEvent):

    js_name = 'Layertree.layerPainted'
    hashable = ['layerId']
    is_hashable = True

    def __init__(self,
                 layerId: Union['LayerId', dict],
                 clip: Union['DOM.Rect', dict],
                 ):
        if isinstance(layerId, dict):
            layerId = LayerId(**layerId)
        self.layerId = layerId
        if isinstance(clip, dict):
            clip = DOM.Rect(**clip)
        self.clip = clip

    @classmethod
    def build_hash(cls, layerId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class LayerTreeDidChangeEvent(BaseEvent):

    js_name = 'Layertree.layerTreeDidChange'
    hashable = []
    is_hashable = False

    def __init__(self,
                 layers: Union['[Layer]', dict, None] = None,
                 ):
        if isinstance(layers, dict):
            layers = [Layer](**layers)
        self.layers = layers

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')



# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)

# Metric: Run-time execution metric.
class Metric(ChromeTypeBase):
    def __init__(self,
                 name: Union['str'],
                 value: Union['float'],
                 ):

        self.name = name
        self.value = value


class Performance(PayloadMixin):
    """ 
    """
    @classmethod
    def disable(cls):
        """Disable collecting and reporting metrics.
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def enable(cls):
        """Enable collecting and reporting metrics.
        """
        return (
            cls.build_send_payload("enable", {
            }),
            None
        )

    @classmethod
    def setTimeDomain(cls,
                      timeDomain: Union['str'],
                      ):
        """Sets time domain to use for collecting and reporting duration metrics.
Note that this must be called before enabling metrics collection. Calling
this method while metrics collection is enabled returns an error.
        :param timeDomain: Time domain
        :type timeDomain: str
        """
        return (
            cls.build_send_payload("setTimeDomain", {
                "timeDomain": timeDomain,
            }),
            None
        )

    @classmethod
    def getMetrics(cls):
        """Retrieve current values of run-time metrics.
        """
        return (
            cls.build_send_payload("getMetrics", {
            }),
            cls.convert_payload({
                "metrics": {
                    "class": [Metric],
                    "optional": False
                },
            })
        )



class MetricsEvent(BaseEvent):

    js_name = 'Performance.metrics'
    hashable = []
    is_hashable = False

    def __init__(self,
                 metrics: Union['[Metric]', dict],
                 title: Union['str', dict],
                 ):
        if isinstance(metrics, dict):
            metrics = [Metric](**metrics)
        self.metrics = metrics
        if isinstance(title, dict):
            title = str(**title)
        self.title = title

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')



import asyncio.subprocess
import logging
import logging.config
import platform
import signal
import os
from collections import namedtuple
import time

from aiohttp import web
import yaml

from chromewhip.chrome import Chrome
from chromewhip.middleware import error_middleware
from chromewhip.routes import setup_routes


log = logging.getLogger(__name__)

HOST = '127.0.0.1'
PORT = 9222
NUM_TABS = 4
DISPLAY = ':99'

async def on_shutdown(app):
    c = app['chrome-driver']
    if c.is_connected:
        for tab in c.tabs:
            await tab.disconnect()

    chrome = app['chrome-process']
    chrome.send_signal(signal.SIGINT)
    try:
        returncode = await asyncio.wait_for(chrome.wait(), timeout=15)
        if not returncode:
            log.error('Timed out trying to shutdown Chrome gracefully!')
        elif returncode < 0:
            log.error('Error code "%s" received while shutting down Chrome!' % abs(returncode))
        else:
            log.debug("Successfully shut down Chrome!")
    except asyncio.TimeoutError:
        log.error('Timed out trying to shutdown Chrome gracefully!')

Settings = namedtuple('Settings', [
    'chrome_fp',
    'chrome_flags',
    'should_run_xfvb'
])

def get_settings():
    chrome_flags = [
        '--window-size=1920,1080',
        '--enable-logging',
        '--hide-scrollbars',
        '--no-first-run',
        '--remote-debugging-address=%s' % HOST,
        '--remote-debugging-port=%s' % PORT,
        '--user-data-dir=/tmp',
        'about:blank'  # TODO: multiple tabs
    ]
    os_type = platform.system()
    if os_type == 'Linux':
        chrome_flags.insert(3, '--no-sandbox')
        chrome_fp = '/opt/google/chrome/chrome'
        should_run_xfvb = True
    elif os_type == 'Darwin':
        chrome_fp = '/Applications/Google Chrome Canary.app/Contents/MacOS/Google Chrome Canary'
        should_run_xfvb = False
    else:
        raise Exception('"%s" system is not supported!' % os_type)
    return Settings(
        chrome_fp,
        chrome_flags,
        should_run_xfvb
    )


def setup_chrome(settings: Settings, env: dict = None, loop: asyncio.AbstractEventLoop = None):
    # TODO: manage process lifecycle in coro
    args = [settings.chrome_fp] + settings.chrome_flags
    chrome = asyncio.subprocess.create_subprocess_exec(*args, env=env, loop=loop)
    return chrome


def setup_xvfb(settings: Settings, env: dict = None, loop: asyncio.AbstractEventLoop = None):
    # TODO: manage process lifecycle in coro
    if not settings.should_run_xfvb:
        return
    flags = [
        DISPLAY,
        '-ac',
        '-screen',
        '0',
        '1920x1080x16',
        '+extension',
        'RANDR',
        '-nolisten',
        'tcp',
    ]
    args = ['/usr/bin/Xvfb'] + flags
    xvfb = asyncio.subprocess.create_subprocess_exec(*args, env=env, loop=loop)
    return xvfb


def setup_app(loop=None, js_profiles_path=None):
    app = web.Application(loop=loop, middlewares=[error_middleware])

    js_profiles = {}

    if js_profiles_path:
        root, _, files, _ = next(os.fwalk(js_profiles_path))
        js_files = filter(lambda f: os.path.splitext(f)[1] == '.js', files)
        _, profile_name = os.path.split(root)
        log.debug('adding profile "{}"'.format(profile_name))
        js_profiles[profile_name] = ""
        for f in js_files:
            code = open(os.path.join(root, f)).read()
            js_profiles[profile_name] += '{}\n'.format(code)

    app.on_shutdown.append(on_shutdown)

    c = Chrome(host=HOST, port=PORT)

    app['chrome-driver'] = c
    app['js-profiles'] = js_profiles

    setup_routes(app)

    return app


if __name__ == '__main__':
    import argparse
    import sys
    config_fp = os.path.abspath(os.path.join(os.path.dirname(__file__), '../config/dev.yaml'))
    config_f = open(config_fp)
    config = yaml.load(config_f)
    logging.config.dictConfig(config['logging'])
    parser = argparse.ArgumentParser()
    parser.add_argument('--js-profiles-path',
                        help="path to a folder with javascript profiles")
    args = parser.parse_args(sys.argv[1:])
    kwargs = {}
    if args.js_profiles_path:
        kwargs['js_profiles_path'] = args.js_profiles_path

    loop = asyncio.get_event_loop()

    env = {
       'DISPLAY': DISPLAY
    }

    settings = get_settings()
    app = setup_app(**kwargs, loop=loop)

    if settings.should_run_xfvb:
        xvfb = setup_xvfb(settings, env=env, loop=loop)
        xvfb_future = loop.run_until_complete(xvfb)
        log.debug('Started xvfb!')
        app['xvfb-process'] = xvfb_future

    chrome = setup_chrome(settings, env=env, loop=loop)
    chrome_future = loop.run_until_complete(chrome)
    time.sleep(3)  # TODO: use event for continuing as opposed to sleep

    log.debug('Started Chrome!')
    app['chrome-process'] = chrome_future

    # TODO: need indication from chrome process to start http server
    loop.run_until_complete(asyncio.sleep(3))
    loop.run_until_complete(app['chrome-driver'].connect())
    web.run_app(app)
